'''
email parser module, version 0.6.3

(c) 2003-2004 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>,
              A'rpi <arpi(at)mplayerhq.hu>

Some parts are based on python's email package written by:
              Ben Gertzfield <che(at)debian.org>
              Barry Warsaw <barry(at)zope.com>
                                                                                
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
'''

from avlib import *
from match import match_any
import os,re
import mimetypes
import binascii,uu,base64,binhex,quopri
from StringIO import StringIO

__all__=['parsemail','attach_name']

def hdr_parse(s,pmode=1):
    params=[]
    qflag=iflag=icnt=cflag=0
    comment=key=value=""
    keymode=1
    lastnonws=0
    wsflag=1
    for c in s:
	if not qflag:
	    # process control chars, by priority:
	    if c=='\\':
		qflag=1
		continue;
	    if c=='"':
		iflag=not iflag
		icnt+=1
		wsflag=0
		continue;
	    if not iflag:
		if wsflag:
		    # skip initial whitespaces, unless they're quoted
		    if c in ('\t',' '):
			continue
		    wsflag=0
		if c=='(':
		    cflag+=1
		    continue
		if c==')':
		    cflag-=1
		    continue
		if cflag==0 and pmode:
		    if c=='=' and keymode:
			keymode=0
			lastnonws=0
			wsflag=1
			continue
		    if c==';':
			params.append((key.strip().lower(),value[:lastnonws],comment.strip()))
			comment=""
			key=""
			value=""
			keymode=1
			icnt=0
			continue
	if cflag>0:
	    comment+=c
	else:
	    if keymode:
		key+=c
	    else:
		value+=c
		if iflag or qflag or not c in ('\t',' '):
		    lastnonws=len(value)
	qflag=0

    if not pmode:
	return (key.strip(),comment.strip())
    params.append((key.strip().lower(),value[:lastnonws],comment.strip()))
    return params

# inspired by Header.py::__unicode__()
ecre = re.compile(r'''
  =\?                   # literal =?
  (?P<charset>[^?]*?)   # non-greedy up to the next ? is the charset
  \?                    # literal ?
  (?P<encoding>[qb])    # either a "q" or a "b", case insensitive
  \?                    # literal ?
  (?P<encoded>.*?)      # non-greedy up to the next ?= is the encoded string
  \?=                   # literal ?=
  ''', re.VERBOSE | re.IGNORECASE)

def decode_header(header):
    """Decode a message header value without converting charset.

    Returns a list of (decoded_string, charset) pairs containing each of the
    decoded parts of the header.  Charset is None for non-encoded parts of the
    header, otherwise a lower-case string containing the name of the character
    set specified in the encoded string.

    An email.Errors.HeaderParseError may be raised when certain decoding error
    occurs (e.g. a base64 decoding exception).
    """
    # If no encoding, just return the header
    header = str(header)
    if not ecre.search(header):
        return [(header, None)]
    decoded = []
    dec = ''
    for line in header.splitlines():
        # This line might not have an encoding in it
        if not ecre.search(line):
            decoded.append((line, None))
            continue
        parts = ecre.split(line)
        while parts:
            unenc = parts.pop(0).strip()
            if unenc:
                # Should we continue a long line?
                if decoded and decoded[-1][1] is None:
                    decoded[-1] = (decoded[-1][0] + SPACE + unenc, None)
                else:
                    decoded.append((unenc, None))
            if parts:
                charset, encoding = [s.lower() for s in parts[0:2]]
                encoded = parts[2]
                dec = None
                if encoding == 'q':
                    dec = quopri.decodestring(encoded)
                elif encoding == 'b':
                    try:
                        dec = base64.decodestring(encoded)
                    except binascii.Error:
                        raise HeaderParseError
                    except ValueError:
                        raise HeaderParseError
                if dec is None:
                    dec = encoded
                if decoded and decoded[-1][1] == charset:
                    decoded[-1] = (decoded[-1][0] + dec, decoded[-1][1])
                else:
                    decoded.append((dec, charset))
            del parts[0:3]
    return decoded

def hdr_decode(h):
    if not h:
	return u""
    uchunks = []
    last = None
    for s,enc in decode_header(h):
        if last not in (None, 'us-ascii'):
            if enc in (None, 'us-ascii'):
                uchunks.append(u' ')
        else:
	    if enc not in (None, 'us-ascii'):
        	uchunks.append(u' ')
	last=enc
	try:
	    uchunks.append(unicode(s,enc or "latin-1"))
	except:
	    uchunks.append(unicode(s,"latin-1"))
    return u''.join(uchunks)

def plain(s):
    return s

class decode_email:
  def debug(self,level,str):
      if debug.debug_level<5:
        self.max_errors-=1
      if self.max_errors==0:
        debug.echo(level,"parsemail(): TOO MANY ERRORS !!!")
      elif self.max_errors>0:
        debug.echo(level,str)
  def scan_part(self,filename=None):
      stream=self.part.getvalue()
      if not stream:
        return
      if filename:
        self.filename=normalize_filename(filename)
        self.files.append(self.filename)
      else:
        self.filename='unknown.bin'
      for scanner in self.scanners:
        scanner.filename=self.filename
      debug.echo(5,"parsemail(): stream len=",len(stream),
                   ", filename=",self.filename)
      debug.echo(9,"decoded data part: "+str([stream[0:40]]),"...")
      for scanner in self.scanners:
        if scanner.is_multistream:
          scanner.addstream(stream)
        else:
          scanner.prescan()
          l,v,r=scanner.scanstream(stream)
          scanner.postscan(l,v,r)
          if not self.vir:
            self.vir=v
          self.ret+=r
          self.level+=l
      # delete old part and create a new StringIO
      self.part=StringIO()
  def __init__(self,data,scanners):
      '''
      Decode whole email and run a scanners on decoded parts.
      '''
      self.level,self.vir,self.ret,self.files=0.0,'',[],[]
      self.scanners=scanners
      in_header=1
      bounds={}
      header=[]
      self.lineno=0
      preline=''
      self.part=StringIO()
      file_name=''
      dio=StringIO(data)
      self.max_errors=20
      while 1:
        line=dio.readline()
        if not line: break
        line=line.rstrip("\r\n")
        self.lineno+=1
        while line[0:2]=="> ":
          line=line[2:]
        if in_header:
          # if line is (really) empty => EOH
          if not line:
            self.scan_part(file_name)
            file_type,file_name,coding=None,None,None
            try:
              p_ctype=p_cdisp=p_ctenc=None
              for hs in header:
                fpos=hs.find(':')
                if fpos<=0:
                  # not a field, check if it's a boundary!
                  b_ok=0
                  for b in bounds.keys():
                    if hs.find("--"+b)==0:
                      debug.echo(8,"HDR: found a boundary inside a header!")
                      b_ok=1
                      break
                  if not b_ok:
                    self.debug(1,"HDR: Invalid header field: '"+hs+"'")
                  continue
                if len(hs)>998:
                  self.debug(1,"HDR: Header line too long! (%d chars)" % (len(hs)))
                field=hs[0:fpos].strip().lower()
                fval=hs[fpos+1:]
                # parse content-type: header
                if field=="content-type":
                  if p_ctype:
                    self.debug(1,"HDR: MIME error: content-type redefined!")
                  p_ctype=hdr_parse(fval)
                  file_type=p_ctype[0][0]
                  for param in p_ctype:
                    if param[0]=="name":
                      if not file_name:
                        file_name=hdr_decode(param[1])
                    elif param[0]=="boundary":
                      # not sure if we have to strip() it:
                      bounds[param[1].strip()]=1
                  file_name=mimetypes.guess_extension(file_type)
                  if file_name:
                    file_name="unknown"+file_name
                  continue
                # parse content-disposition: header
                if field=="content-disposition":
                  if p_cdisp:
                    self.debug(1,"HDR: MIME error: content-disposition redefined!")
                  p_cdisp=hdr_parse(fval)
                  for param in p_cdisp:
                    if param[0] in ["filename"]:
                      file_name=hdr_decode(param[1])
                  continue
                if field=="content-transfer-encoding":
                  if p_ctenc:
                    self.debug(1,"HDR: MIME error: transfer-encoding redefined!")
                  else:
                    coding=hdr_parse(fval,0)[0]
                  continue
            except:
              self.debug(1,"HDR: Exception while MIME header parsing")
              if self.max_errors>0:
                debug.traceback(4,"HDR: ")
            decoder=plain
            if coding:
              coding=coding.strip().lower()
              if coding=="quoted-printable":
                decoder=binascii.a2b_qp
              elif coding=="base64":
                decoder=binascii.a2b_base64
              elif coding in ('x-uuencode', 'uuencode', 'uue', 'x-uue'):
                decoder=binascii.a2b_uu
              elif file_type=="application/mac-binhex40":
                decoder=binascii.a2b_hqx
            checkend=0
            in_header=0
            header=[]
            self.part=StringIO() # create new part
            continue
        else: # not in_header
          # check for boundaries
          for b in bounds.keys():
            if line.find("--"+b)>=0:
              in_header=1
              debug.echo(8,"HDR: boundary found, switching to in_header")
              continue
          # handle forwarded/bounced MIME emails
          lline=line.strip().lower()
          fpos=lline.find(':')
          if fpos>0:
            field=lline[0:fpos]
            if field in ("content-type","content-transfer-encoding","content-disposition","content-id"):
              in_header=1
              debug.echo(7,"HDR: looks like header, switching to in_header")
          if lline[0:40]=="(this file must be converted with binhex":
            decoder=binascii.a2b_hqx
            continue
          if lline[0:6]=="begin " or lline[0:13]=="begin-base64 ":
            try:
              begin,mode,name=lline.split(' ',2)
              if int(mode,8)>0 and name:
                file_name=name.strip()
              if begin.lower()=="begin":
                decoder=binascii.a2b_uu
              else:
                decoder=binascii.a2b_base64
              self.scan_part(file_name)
              self.part=StringIO() # create new part
              checkend=1
              continue
            except:
              pass
        if in_header:
          # do the header un-folding:
          if line[0:1] in [' ','\t']:
            if header:
              header[-1]+=line
            else:
              self.debug(1,"syntax error: bad unfolded header line: '"+line+"'")
          else:
            header.append(line)
        else:
          # decode
          try:
            preline+=line
            if preline:
              self.part.write(decoder(preline))
              preline=''
          except:
            if checkend and line.lower()[0:3]=="end":
              self.scan_part(file_name)
              decoder=None
              self.part=StringIO()
              checkend=0
              preline=''
              continue
            if line[-2:]=='=9':
              # correct end
              preline=''
              continue
            self.debug(3,
              "parsemail(): Exception while attachment decoding, file may be truncated! length=%d, line[%d]=...'%s'" \
                % (len(preline),self.lineno,preline[-80:]))
      self.scan_part(file_name)
  def scanstream(self):
      for scanner in self.scanners:
        if scanner.is_multistream:
          debug.echo(5,"Scanning as multistream")
          self.level,self.vir,self.ret=scanner.scanmultistream([])
          if is_infected(self.level,self.vir):
            return self.level,self.vir,self.ret
        del scanner
      return self.level,self.vir,self.ret

class parsemail(match_any):
  '''
  mimeparse interscanner
  
  Usage: parsemail(scanner1(...) [,scanner2(...)] )
  
  This scanner parses emails and send multiple of streams into
  scanner1(). If an multiscanner is used as another_scanner,
  optimal method is used (scans all files at once).
  '''
  name='ParseMail()'
  def __init__(self,*scanners):
      #self.name="ParseMail("+scanner.name+")"
      match_any.__init__(self,scanners)
  def scanstream(self,stream):
      return decode_email(stream,self.scanners).scanstream()

class attach_name(ascanner):
  '''
  Attachment name scanner.
  
  Usage: attach_name({'VirName':'regexp_pattern',..},flags=re.I)
  
  Where: 'VirName' is a string, which defines virus name
         'regexp_pattern' is an regular expression. If this expression
           is found in attachment's name, VirName is returned
           as virus name.
         flags is a number, which defines regular expression flags.
           By default IGNORECASE is used.
  
  This scanner checks defined name of file in mime attachments.
  By default known executable extensions will be blocked.
  
  This scanner can be used only if it's parent is parsemail().
  '''
  name='attach_name()'
  def __init__(self,wrong_types={'Executable':'\.(exe|com|pif|lnk|scr|vbs|cpl|vbs|js)$'},flags=re.I):
      # compile regexps
      if type(wrong_types)==type(''):
        self.wrong_regs={'BLOCKED_NAME':re.compile(wrong_types,flags)}
      else:
        self.wrong_regs={}
        for vname,reg in wrong_types.items():
          self.wrong_regs[vname]=re.compile(reg,flags)
  def scanstream(self,stream):
      debug.echo(7,"attach_name(): filename: ",self.filename)
      for vname,reg in self.wrong_regs.items():
        if reg.search(self.filename):
          return iret(1.0,vname,[self.name+": "+vname+' found in '+self.filename+'!\n'])
      return 0.0,'',[]



#!/usr/bin/python

'''
sagator.py
(c) 2003-2004 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

'''

from aglib import *
import signal,time
import avlib

# load config
from default import *
try:
  from config import *
except ImportError,es:
  if str(es)[-6:]=="config":
    print "ERROR! Config file not found! Exiting now."
    sys.exit(1)
  else:
    raise

if __name__ == '__main__':
  debug.set_level(DEBUG_LEVEL)
  n=1
  while n<len(sys.argv):
    if sys.argv[n]=='--help':
      print "SAGATOR",SG_VER_REL
      print "(c) Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>"
      print "Licensed under GNU GPL."
      print ""
      print "Params: --help        this help"
      print "        --config f    load \"f\" as alternate config (without .py extension)"
      print "        --nodaemon    don't daemonize after startup"
      print "        --logfile l   filename for logging ('-' for stdout)"
      print "        --debug l     set debug level to l"
      sys.exit(0)
    elif sys.argv[n]=='--config':
      try:
        exec("from %s import *" % sys.argv[n+1])
      except ImportError,es:
        print "ImportError:",es
      n+=2
    elif sys.argv[n]=='--daemon':
      globals.daemon=True
      n+=1
    elif sys.argv[n] in ('--nodaemon','-n'):
      globals.daemon=False
      n+=1
    elif sys.argv[n]=='--logfile':
      LOGFILE=sys.argv[n+1]
      n+=2
    elif sys.argv[n]=='--debug':
      debug.set_level(int(sys.argv[n+1]))
      n+=2
    else:
      print "Unknown parameter:", sys.argv[n]
      sys.exit(1)
  debug.set_logfile(LOGFILE)
  debug.chroot=CHROOT
  globals.USER=USER
  globals.GROUP=GROUP
  globals.CHROOT=CHROOT
  globals.pid_file=PID_FILE
  avlib.smtp.SMTP_SERVER=SMTP_SERVER
  signal.signal(signal.SIGUSR2,sigusr2)
  if globals.daemon:
    os.close(0) # close stdin
  debug.echo(1,"SAGATOR %s starting at %s" \
    % (SG_VER_REL,time.strftime("%c",time.localtime())))
  try:
    # open PID file
    try:
      if globals.daemon:
        globals.pidf=open(globals.pid_file,'w')
    except:
      if globals.pid_file:
        debug.echo(0,"Can't open pid file: ",globals.pid_file)
    # start services
    for service in SRV:
      pids=service.main()
      globals.service_pids.extend(pids)
    for pid in globals.service_pids:
      if globals.daemon:
        globals.pidf.write(str(pid)+" ")
    debug.echo(6,"Services finished, pids: ",str(globals.service_pids))
    if not globals.daemon:
      while True:
        time.sleep(60)
  except KeyboardInterrupt:
    sigterm('KeyboardInterrupt') # kill self



'''
Example realscanners for sagator, version 0.1

(c) 2004 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
    
'''

from avlib import *
import re,string

__all__=['av_example']

class av_example(ascanner):
  '''
  av_example() realscanner can be used as template for your realscanner
  
  Usage: av_example(args...)
  '''
  
  # Define scanner name here. A realscanner's name must end with ().
  name='av_example()'

  def __init__(self,args=None):
      # This function is called from sagator's config file.
      # You can store it's parameters here, but do not make
      # other things.
      self.args=args
  def scanstream(self,stream):
      # This function is called, when an scan over a buffer is required.
      # Buffer is passed as "stream" parameter.
      # Return values:
      #   an integer - level of virus. >=1.0 means, that virus has been found
      #   an string - a simple virus name of first virus found
      #   an array - lines returned by this scanner (used in reports
      #              and for debugging

      # You need to check, if it contains a virus. For example
      # we check, if it contains an character "x".
      if 'x' in stream:
        # now a virus (char 'x') has been found. Returning it:
        return 1.0,'Char:x',["Char 'x' is found!"]
      else:
        # these data are clean
        return 0.0,'',[]
      # if an error occured, you also can raise an error
      raise ScannerError,'Example error'
  def scanfile(self,files,dir=''):
      # This function is similiar to scanstream().
      # files contain an array of files returned to scan.
      # These files are in 'dir' directory.
      # You have 2 choices to scan these files. You can scan
      # files from "files" array or you can scan whole 'dir'.
      # Return values are same as in scanstream().
      pass



# Antivir scanners

from basic import *
from dazuko import *
from clamav import *
from avg import avgd
from bdc import bdc
from nod import *
from sophie import sophie
from kav import *
from icap import *
from sanitize import *
from filter import *
from cmd import *
#from example import *



'''
Sagators statistics collector and distributor, version 0.3.2

(c) 2003-2004 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>
                                                                                
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

Examples:
  python2 sagator.py /
  python2 sagator.py mrtg Total-count Filtered-count

'''

import socket,re,string,os,sys
import urllib,time,signal,struct
import interscan.match
from avlib import *

__all__ = ['collector','update','query','status']

COLLECTOR_SERVER=()
MYCOLLECTOR=None

def uptime(starttime):
    up=time.mktime(time.localtime())-starttime
    mins=int(up/60)%60
    hours=int(up/3600)%24
    days=int(up/86400)
    if days>0:
      return "%d days, %02d:%02d" % (days,hours,mins)
    else:
      return "%02d:%02d" % (hours,mins)

class collector:
  pid=0
  data={ 'Total-count':0.0, 'Total-bytes':0.0,
         'Filtered-count':0.0, 'Filtered-bytes':0.0,
         'Total-time':0.0 }
  def __init__(self,bindaddr,statusfile='/tmp/status'):
      global MYCOLLECTOR
      MYCOLLECTOR=self
      # startup initialization
      self.bindaddr=bindaddr
      self.statusfile=statusfile
      # create a server
      self.s=socket.socket(socket.AF_INET,socket.SOCK_STREAM)
      self.s.setsockopt(socket.SOL_SOCKET,socket.SO_REUSEADDR,1)
      self.s.bind(self.bindaddr)
      self.s.listen(10)
  def main(self):
      global COLLECTOR_SERVER
      COLLECTOR_SERVER=self.bindaddr
      self.lastsave=time.mktime(time.localtime())
      self.pipe_r,self.pipe_w=os.pipe() # pipe for updates
      self.pid=os.fork()
      if self.pid==0: # main process
        # chroot
        from aglib import dochroot,sighup
        dochroot()
        signal.signal(signal.SIGHUP,sighup)
        signal.signal(signal.SIGUSR1,self.sigusr1)
        signal.signal(signal.SIGTERM,self.sigterm)
        if globals.daemon:
          globals.pidf.close()
        os.close(self.pipe_w)
        debug.dup()
        self.data['Start-at']=self.lastsave
        self.loadstats()
        self.data['Restart-at']=self.lastsave
        self.data['Hostname']=socket.gethostname()
        debug.echo(1,"collector(): service started, waiting for connections...")
        self.mypid=os.getpid()
        while 1:
          try:
            self.conn,addr=self.s.accept()
            socket_settimeout(self.conn,8)
            try:
              f=self.conn.makefile("rw",0)
              req=f.readline()
              debug.echo(5,"collector(): QUERY: ",req.strip())
              proto=0.0
              reg1=re.search("^GET /([^ ]*) HTTP/(1.[01])",req)
              if reg1:
                dirs=reg1.group(1).split("/")
                proto=float(reg1.group(2))
              else:
                reg1=re.search("^GET /([^ ]*)$",req.rstrip())
                if reg1:
                  proto=1.0
                  dirs=reg1.group(1).split("/")
              if proto>=1.0:
                args=self.parseargs(dirs[1:])
                lines={}
                while proto>=1.1:
                  line=f.readline()
                  if (line=="\r\n") or (line=="\n"):
                    break
                  a=line.split(": ")
                  if len(a)>1:
                    lines[a[0]]=a[1].strip()
                  else:
                    lines[a[0]]=a[0].strip()
                self.conn.sendall("HTTP/1.1 200 OK\r\n")
                self.conn.sendall("Content-Type: text/plain; charset=iso-8859-1\r\n")
                self.conn.sendall("\r\n");
                self.data["Uptime"]=uptime(self.data['Start-at'])
                self.data["Restart-age"]=uptime(self.data['Restart-at'])
                sstr=''
                if dirs[0]=="":
                  for key,value in self.data.items():
                    sstr+=key+":\t"+str(value)+"\r\n"
                  self.conn.sendall(sstr)
                elif dirs[0]=="mrtg":
                  for key in dirs[1:3]+['Uptime','Hostname']:
                    if key in self.data.keys():
                      try:
                        o="%-20.0f" % self.data[key]
                        sstr+=o.strip()+"\r\n"
                      except TypeError:
                        sstr+=str(self.data[key])+"\r\n"
                    else: # return 0 for unknown key
                      sstr+="0\r\n"
                  self.conn.sendall(sstr)
                else:
                  self.message("404 Not Found")
              else:
                self.message("501 Method Not Implemented")
                self.conn.sendall(req)
            except socket.error,(ec,es):
              if not ec in [32,104,107]:
                debug.echo(3,"collector(): socket.error: ",ec,es)
                debug.traceback(4,"collector(): ")
            except:
              debug.traceback(3,"collector(): ")
            try:
              self.conn.shutdown(2)
              self.conn.close()
            except socket.error:
              pass 
          except:
            try:
              self.conn.shutdown(2)
              self.conn.close()
            except:
              pass
        # while 1 ... end
      else:
        self.s.close()
        os.close(self.pipe_r)
      return self.pid
  def update(self,dirs):
      d=string.join(dirs,"/")+"\n"+' '*1024
      signal.signal(signal.SIGALRM,self.sigalrm)
      try:
        signal.alarm(2)
        os.write(self.pipe_w,d[:1024])
        signal.alarm(0)
      except OSError,(ec,es):
        signal.alarm(0)
        if ec==4: # Interrupted system call
          debug.echo(3,"collector(): writing to pipe: ",es)
      except:
        debug.echo(1,"collector(): unknown error on update")
        debug.traceback(3,"collector()")
      signal.signal(signal.SIGALRM,signal.SIG_DFL)
      os.kill(self.pid,signal.SIGUSR1)
  def sigalrm(self,sn,stack):
      # ignore this signal, only do an "interrupted system call"
      debug.echo(1,"collector(): WARNING: Interrupting write to pipe! (pipe full?)")
  def sigusr1(self,sn,stack):
      signal.signal(signal.SIGALRM,self.sigalrm)
      try:
        signal.alarm(2)
        sstr=os.read(self.pipe_r,1024)
        signal.alarm(0)
      except OSError,(ec,es):
        signal.alarm(0)
        if ec==4: # Interrupted system call
          debug.echo(3,"collector(): ",es)
          return
      signal.signal(signal.SIGALRM,signal.SIG_DFL)
      args=self.parseargs(sstr.split("/"))
      self.data["Childrens"]=0.0
      for k,v in args.items():
        try:
          self.data[k]+=float(v)
        except KeyError:
          self.data[k]=float(v)
        except ValueError:
          self.data[k]=v
        except TypeError:
          debug.echo(0,"collector(): TypeError: ",k,v,type(v))
      # save stats evry 5 minutes
      ctime=time.mktime(time.localtime())
      if (ctime-self.lastsave)>300:
        self.lastsave=ctime
        self.savestats()
  def sigterm(self,sn,stack):
      try:
        os.close(self.pipe_r)
        debug.echo(4,"collector(): Saving stats...")
        self.savestats()
      except:
        pass
      debug.echo(1,"collector(): Exiting - SIGTERM ...")
      os._exit(0)
  def loadstats(self):
      try:
        f=open(self.statusfile,'r')
        for line in f.readlines():
          a=line.split(":\t")
          try:
            self.data[a[0]]=float(a[1])
          except ValueError:
            self.data[a[0]]=a[1]
          except IndexError:
            # ignore wrong lines from status file
            pass
        f.close()
      except IOError,(ec,es):
        debug.echo(1,"collector(): loadstat error: ",es)
      except:
        debug.echo(1,"collector(): Can't load status file.")
        debug.traceback(1,"collector()")
  def savestats(self):
      f=open(self.statusfile,'w')
      for k,v in self.data.items():
        f.write(k+":\t"+str(v)+"\n")
      f.close()
      debug.echo(4,"collector(): Status saved.")
  def message(self,sstr):
      self.conn.sendall("HTTP/1.1 "+sstr+"\r\n\r\n")
  def parseargs(self,dirs):
      args={}
      for arg in dirs:
        a=arg.split("=")
        k=urllib.unquote_plus(a[0])
        if re.search("^[A-Za-z0-9-]*$",k):
          if len(a)>1:
            args[k]=urllib.unquote_plus(a[1])
          else:
            args[k]='1'
        else:
          debug.echo(1,"collector(): Unknown key: ",k)
      return args

def query(qt,args):
    if COLLECTOR_SERVER:
      s=socket.socket(socket.AF_INET,socket.SOCK_STREAM)
      socket_settimeout(s,10)
      s.connect(COLLECTOR_SERVER)
      urlargv=''
      for arg in args:
        urlargv+="/"+urllib.quote_plus(arg,"/=")
      s.sendall("GET /"+qt+urlargv+" HTTP/1.1\r\n\r\n")
      f=s.makefile("r",0)
      lines=[]
      for _ in range(512):
        line=f.readline()
        if (line=="\r\n") or (line=="\n") or (line==""):
          break;
        lines.append(line)
      lines=[]
      s.shutdown(2)
      return s.recv(9999).rstrip()
    else:
      return None

def update(bs=-1,t=0,infected=0,tempfail=0):
    '''update statistics by 'bs' bytes, 't' time,
       if infected, then update also filtered bytes,files
    '''
    args=['Total-count','Total-bytes='+str(bs),'Total-time='+str(t),
          'Temp-fail='+str(tempfail),'Childrens=0']
    if infected:
      args.extend(['Filtered-count','Filtered-bytes='+str(bs)])
    try:
      MYCOLLECTOR.update(args+status.collect)
      status.collect=[]
    except:
      pass

class rrd:
  RRDDATA={
    'Total-bytes':		'totalbs',
    'Filtered-bytes':		'filteredbs',
    'Virus-bytes':		'virusbs',
    'Spam-bytes':		'spambs',
    'Total-count':		'total',
    'Filtered-count':		'filtered',
    'Virus-count':		'virus',
    'Spam-count':		'spam',
    'Temp-fail':		'fails',
    'Total-time':		'time'
  }
  DATES=[
    [3600*24,		'Day',		1,	600],
    [3600*24*7,		'Week',		6,	700],
    [3600*24*31,	'Month',	24,	775],
    [3600*24*365,	'Year',		288,	797]
  ]
  PARAMS='''--imgformat PNG --interlaced --width 576'''
  def __init__(self,path='./'):
      self.RRDPATH=path
      self.RRDFILE=os.path.join(path,'sagator.rrd')
  def create(self,xparams=''):
      # create an RRD
      cmd='rrdtool create %s %s' % (self.RRDFILE,xparams)
      for ds in self.RRDDATA.values():
        cmd+=' DS:%s:COUNTER:600:0:U' % ds
      for p1 in ['AVERAGE','MAX']:
        for p2 in self.DATES:
          cmd+=' RRA:%s:0.5:%d:%d' % (p1,p2[2],p2[3])
      os.system(cmd)
      return cmd
  def update(self):
      try:
        open(self.RRDFILE)
      except IOError:
        self.create()
      q=query('/',[])
      cmd='rrdtool update %s N' % self.RRDFILE
      for key,value in self.RRDDATA.items():
        reg1=re.compile('^%s:[ \t]*([0-9]+)' % key,re.M).search(q)
        if reg1:
          cmd+=':'+reg1.group(1)
        else:
          cmd+=':0'
          #print "WARNING: Key %s not found!" % key
      os.system(cmd)
      return self
  def system(self,s):
      os.system(s.replace('$STEP',self.STEP).replace('$PATH',self.RRDPATH))
  def range(self,delta,r,xparams):
      self.system('''
        rrdtool graph $PATH/%s-count.png %s %s \
        --vertical-label msgs/min --units-exponent 0 --start -%s \
        DEF:total=$PATH/sagator.rrd:total:AVERAGE \
        DEF:filtered=$PATH/sagator.rrd:filtered:AVERAGE \
        DEF:virus=$PATH/sagator.rrd:virus:AVERAGE \
        DEF:spam=$PATH/sagator.rrd:spam:AVERAGE \
        CDEF:mtotal=total,60,* \
        CDEF:mfiltered=filtered,60,* \
        CDEF:mvirus=virus,60,* \
        CDEF:mspam=spam,60,* \
        CDEF:dtotal=total,UN,0,total,IF,$STEP,* \
        CDEF:stotal=PREV,UN,dtotal,PREV,IF,dtotal,+ \
        CDEF:dfiltered=filtered,UN,0,filtered,IF,$STEP,* \
        CDEF:sfiltered=PREV,UN,dfiltered,PREV,IF,dfiltered,+ \
        CDEF:dvirus=virus,UN,0,virus,IF,$STEP,* \
        CDEF:svirus=PREV,UN,dvirus,PREV,IF,dvirus,+ \
        CDEF:dspam=spam,UN,0,spam,IF,$STEP,* \
        CDEF:sspam=PREV,UN,dspam,PREV,IF,dspam,+ \
        CDEF:rfiltered=sfiltered,100,*,stotal,/ \
        CDEF:rvirus=svirus,100,*,stotal,/ \
        CDEF:rspam=sspam,100,*,stotal,/ \
        AREA:mtotal#00a000:"Total   " \
        GPRINT:stotal:MAX:"%%8.0lf msgs" \
        GPRINT:mtotal:AVERAGE:"avg\\: %%5.2lf%%s/min" \
        GPRINT:mtotal:MAX:"max\\: %%5.2lf%%s/min\\l" \
        LINE2:mfiltered#c0c000:"Filtered" \
        GPRINT:sfiltered:MAX:"%%8.0lf msgs" \
        GPRINT:rfiltered:AVERAGE:"(%%6.2lf %%%%)\\l" \
        LINE2:mvirus#a00000:"Virus   " \
        GPRINT:svirus:MAX:"%%8.0lf msgs" \
        GPRINT:rvirus:AVERAGE:"(%%6.2lf %%%%)\\l" \
        LINE2:mspam#0000a0:"Spam    " \
        GPRINT:sspam:MAX:"%%8.0lf msgs" \
        GPRINT:rspam:AVERAGE:"(%%6.2lf %%%%)" \
        COMMENT:"                    Generated at %s by SAGATOR"
      ''' % (delta,self.PARAMS,xparams,r,time.strftime("%c")))
      self.system('''
        rrdtool graph $PATH/%s-bytes.png %s %s \
        --vertical-label bytes/min --start -%s \
        DEF:total=$PATH/sagator.rrd:totalbs:AVERAGE \
        DEF:filtered=$PATH/sagator.rrd:filteredbs:AVERAGE \
        DEF:virus=$PATH/sagator.rrd:virusbs:AVERAGE \
        DEF:spam=$PATH/sagator.rrd:spambs:AVERAGE \
        CDEF:mtotal=total,60,* \
        CDEF:mfiltered=filtered,60,* \
        CDEF:mvirus=virus,60,* \
        CDEF:mspam=spam,60,* \
        CDEF:dtotal=total,UN,0,total,IF,$STEP,* \
        CDEF:stotal=PREV,UN,dtotal,PREV,IF,dtotal,+ \
        CDEF:dfiltered=filtered,UN,0,filtered,IF,$STEP,* \
        CDEF:sfiltered=PREV,UN,dfiltered,PREV,IF,dfiltered,+ \
        CDEF:dvirus=virus,UN,0,virus,IF,$STEP,* \
        CDEF:svirus=PREV,UN,dvirus,PREV,IF,dvirus,+ \
        CDEF:dspam=spam,UN,0,spam,IF,$STEP,* \
        CDEF:sspam=PREV,UN,dspam,PREV,IF,dspam,+ \
        CDEF:rfiltered=sfiltered,100,*,stotal,/ \
        CDEF:rvirus=svirus,100,*,stotal,/ \
        CDEF:rspam=sspam,100,*,stotal,/ \
        AREA:mtotal#00a000:"Total   " \
        GPRINT:stotal:MAX:"%%8.2lf %%sB" \
        GPRINT:mtotal:AVERAGE:"avg\\: %%5.2lf %%sB/min" \
        GPRINT:mtotal:MAX:"max\\: %%5.2lf %%sB/min\\l" \
        LINE2:mfiltered#c0c000:"Filtered" \
        GPRINT:sfiltered:MAX:"%%8.2lf %%sB" \
        GPRINT:rfiltered:AVERAGE:"(%%6.2lf %%%%)\\l" \
        LINE2:mvirus#a00000:"Virus   " \
        GPRINT:svirus:MAX:"%%8.2lf %%sB" \
        GPRINT:rvirus:AVERAGE:"(%%6.2lf %%%%)\\l" \
        LINE2:mspam#0000a0:"Spam    " \
        GPRINT:sspam:MAX:"%%8.2lf %%sB" \
        GPRINT:rspam:AVERAGE:"(%%6.2lf %%%%)" \
        COMMENT:"                      Generated at %s by SAGATOR"
      ''' % (delta,self.PARAMS,xparams,r,time.strftime("%c")))
      self.system('''
        rrdtool graph $PATH/%s-time.png %s %s \
        --vertical-label "Sagator Load" --units-exponent 0 --start -%s \
        DEF:time=$PATH/sagator.rrd:time:MAX \
        AREA:time#00a000:"Sagator load" \
        GPRINT:time:AVERAGE:"avg\\: %%5.2lf" \
        GPRINT:time:MAX:"max\\: %%5.2lf"
      ''' % (delta,self.PARAMS,xparams,r))
  def graph(self,lazy='--lazy'):
      xparams=''
      for d in self.DATES:
        self.STEP=str(300*d[2])
        self.range(d[1].lower(),d[0],xparams)
        xparams=lazy
      return self

class status(interscan.match.match_any):
  '''
  status() interscanner
  
  Usage: status("String",scanner1(),scanner2(),...)
  
  Where: "String" is a string, which defines a prefix for status update
           This scring can't contain spaces.
         scanner1(),... are sagator's scanners
  
  This interscanner can be used to collect some other statistics.
  Preffered usage:
    SCANNERS=[
      status("Virus",virus_scanner1(),...),
      status("Spam",spam_scanner1(),...)
    ]
  This collects Virus-count,Virus-bytes,Spam-count,Spam-bytes in collector.
  '''
  name='' # making this scanner invisible
  collect=[]
  def __init__(self,string,*scanners):
      self.status_string=string
      interscan.match.match_any.__init__(self,scanners)
  def scanstream(self,stream):
      level,vir,ret=interscan.match.match_any.scanstream(self,stream)
      if vir:
        status.collect.extend([self.status_string+"-count",
          self.status_string+"-bytes="+str(len(stream))])
      return level,vir,ret

if __name__ == '__main__':
  if len(sys.argv)>1:
    # get host and port from collector service
    from config import SRV
    for service in SRV:
      if service.name=='collector()':
        COLLECTOR_SERVER=service.BINDTO
    if sys.argv[1]=='rrdtool':
      rrd(sys.argv[2]).update().graph()
    elif sys.argv[1]=='rrdupdate':
      rrd(sys.argv[2]).update()
    elif sys.argv[1]=='rrdgraph':
      rrd(sys.argv[2]).graph('')
    else:
      print query(sys.argv[1],sys.argv[2:])



'''
NOD module, version 0.6.1

(c) 2003-2004 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>
                                                                                
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

'''

from avlib import *
import os,socket,re,string
from dazuko import dazuko

__all__=['nod2','nod2dazuko','nodold']

class nod2(ascanner):
  '''
  NOD2 scanner. Works with nod32lfs.
  
  Usage: nod2(['command','params...'])

  Where: in [] are command line parameters for nod32

  This scanners is a realscanner, which can be used to scan
  for viruses. You can use nod32 from nod32lfs package.
  '''
  name='nod2()'
  reg_nod2_02=re.compile('^object="[^"]*", name="([^"]*)", virus="([^"]*)", ',
        re.IGNORECASE)
  reg_nod2_01=re.compile(" - (([^ ][^-][^ ]|[a-z/-])*)$",re.IGNORECASE)
  def __init__(self,arg=['/usr/sbin/nod32','--log-brief','--all','--mail','--arch','--adv-heur']):
      self.arg=arg
  def scanfile(self,files,dirname=''):
      level,detected,virlist=0.0,'',[]
      if dirname:
        pf=popen(self.arg+[safe.fn(dirname)])
        pf.tocmd.close()
        for line in pf.readlines():
          virlist.append(line)
        rc=pf.wait()
        pf.fromcmd.close()
        debug.echo(4,"NOD2: ",rc,virlist)
        if rc in [1,10]:
          reg1=self.reg_nod2_02.search(virlist[0])
          if reg1:
            detected=reg1.group(2)
            level+=1.0
          else:
            reg1=self.reg_nod2_01.search(virlist[0])
            level+=1.0
            if reg1:
              detected=reg1.group(1)
              level+=1.0
            else:
              debug.echo(1,"NOD2: Can't decode reply! ",virlist)
              detected="CANT_DECOCE_VIRNAME"
              level+=1.0
      if is_infected(level,detected):
        mail.addheader('X-Sagator-NOD2',detected)
      return level,detected,virlist

class nod2dazuko(dazuko):
  '''
  NOD32lfs scanner over dazuko module.
  
  Usage: nod2dazuko('/dazuko_prefix','logfile')
  
  Where: '/dazuko_prefix' is a string, which defines directory
            and a part of filename to store scanned files,
            by default '/tmp/dazuko'
         'logfile' is a string, which defines full path and filename
            of logfile, by default not logfile used
  '''
  name='nod2dazuko()'
  def __init__(self,dazuko_prefix,logfile=''):
      self.prefix=dazuko_prefix
      self.logfile=logfile
      self.reg=re.compile(
        '^[0-9: -]*: event=[a-z]*, uid=[0-9]*, access=0, object="[^"]*", name="([^"]*)", virus="([^"]*)", ',
        re.IGNORECASE)

class nodold(ascanner):
  '''
  NOD32d scanner. This works only with nod32d from ESET.
  
  Usage: nodoldd(['host',port],params)
  
  Where: 'host' is a string, which defines a host of nod32d
            (by default 'localhost')
         port is a number, which defines a port of nod32d
            (by default 8448)
         params is a string to define scanner parameters.
            It's defaults to "@patt:1@heur:2@clean:0".
            There is no format description.
  
  This scanner mat be used to scan for viruses by ESET's NOD32.
  Only old version (buyed directly from ESET) may be used.
  
  WARNING! NOD32 can't scan direct mboxes and archives.
  You need parsemail() module to parse mailboxes. For exmple:
    mimeparse(stream2file(nodoldd(['localhost',8448])))
  If you need also unpacker support, add decompress interscanner:
    mimeparse(stream2file(decompress(nodoldd(['localhost',8448]))))
  
  If you want start nod32d in chroot, use for example:
    exec chroot /var/spool/vscan /opt/nod32/nod32d
  from any script.
  '''
  name='NodOldD()'
  def __init__(self,arg,params="@patt:1@heur:2@clean:0"):
      self.arg=arg
      self.params=params
  def scanfile(self,filenames,dirname=''):
      level=0.0
      debug.echo(7,"nodold(): ",[filenames],dirname)
      if type(self.arg)==type('string'): # local socket
        s=socket.socket(socket.AF_UNIX,socket.SOCK_STREAM)
        socket_settimeout(s,120)
        s.connect(self.arg)
      else:
        s=socket.socket(socket.AF_INET,socket.SOCK_STREAM)
        socket_settimeout(s,120)
        s.connect((self.arg[0],self.arg[1]))
      f=s.makefile('rw',BUFSIZE)
      r=f.readline() # 200 NOD32SS 1.386 (20030326)
      s.sendall("PARM "+self.params+"\n")
      r=f.readline() # 200 OK (patt: 1, heur: 2, clean: 0)
      ret=[]
      for fname in filenames:
        s.sendall("SCAN "+os.path.join(dirname,fname)+"\n")
        line=f.readline()
        if not re.search("^200",line):
          ret.append(line[4:])
          level+=1.0
        if re.search("^5",line): # error opening file?
          raise ScannerError,'nodold: '+line
      s.sendall("QUIT\n")
      r=f.readline()
      f.close()
      s.shutdown(2)
      s.close()
      if ret==[]:
        detected=''
      else:
        detected=string.rstrip(ret[0])
      if is_infected(level,detected):
        mail.addheader('X-Sagator-NodOld',detected)
      return level,detected,ret



'''
sanitizer module, version 0.1

(c) 2003-2004 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>
                                                                                
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

'''

from avlib import *

__all__=['sanitize']

class sanitize(ascanner):
  '''
  sanitize realscanner
  
  Usage: sanitize([insert_pattern[,extensions_reg_exp]])

  This scanners is a realscanner, which can be used to sanitize
  email. It means all attachments are renamed to safe names.
  
  For example:
    filename.exe will be renamed to filename-exe
  
  WARNING! This scanner is in BETA stage! Id don't conform to RFC 2183!
  '''
  name='sanitize()'
  def __init__(self,insert='-',extensions='[a-z0-9]{3}'):
      self.INS=insert
      re_flags=re.IGNORECASE|re.DOTALL
      self.re_content=re.compile('^ *Content-(Type|Disposition):',re_flags).search
      self.re_nocontent=re.compile('^\S',re_flags).search
      self.re_name1=re.compile('(^.*name="[^"]*)\.('+extensions+'.*$)',re_flags).search
      self.re_name2=re.compile('(^.*name=[^; ]*)\.('+extensions+'[; ].*$)',re_flags).search
  def scanstream(self,stream):
      f=StringIO(mail.data[mail.bodypos:])
      out=StringIO()
      out.write(mail.data[mail.bodypos:])
      incontent=False
      while 1:
        line=f.readline()
        if not line:
          break
        if incontent:
          if self.re_nocontent(line):
            incontent=False
            if self.re_content(line):
              line=self.check(line)
            out.write(line)
            continue
          line=self.check(line)
        else:
          if self.re_content(line):
            incontent=True
            line=self.check(line)
        out.write(line)
      mail.data=out.getvalue()
      # return CLEAN, because sanitizer can't detect a virus
      return 0.0,'',[]
  def check(self,line):
      r1=self.re_name1(line)
      if r1:
        return r1.group(1)+self.INS+r1.group(2)
      r2=self.re_name2(line)
      if r2:
        return r2.group(1)+self.INS+r2.group(2)
      return line



'''
sophie module, version 0.1

(c) 2003 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>
                                                                                
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

'''

from avlib import *
import socket,re,string

__all__=['sophie','trophie']

class sophie(ascanner):
  '''
  sophie realscanner
  
  Usage: sophie(sophie_socket,chroot_path)

  Where: sophie_socket is a filename to sophie via UNIX socket
         chroot_path is a path, which you need to add
           (may be the same as your CHROOT, if you are using
           sagator in chroot and sophie not)

  This scanners is a realscanner, which can be used to scan
  for viruses. Sophie uses sophos libsavi. It can also be used
  for Trophie with TrendMicro antivirus scanner.

  If you need a chrooted sophie, copy it into chroot and start it
  from here. Do not forget to update it's chrooted database.
  '''
  name='sophie()'
  def __init__(self,scanner,chroot_path=''):
      self.scanner=scanner
      self.chroot=chroot_path
  def scanfile(self,files,dir=''):
      level,detected,virlist=0.0,'',[]
      s=socket.socket(socket.AF_UNIX,socket.SOCK_STREAM)
      socket_settimeout(s,120)
      s.connect(self.scanner)
      for fname in files:
        s.sendall(self.chroot+fname+'\n')
        ret=s.recv(256)
        debug.echo(4,self.name+': ',ret)
        if ret!='0':
          reg=re.search('^(-?[0-9]+):(.*)$',ret)
          if reg.group(1)=="1":
            level+=1.0
            virlist.append(fname+': '+reg.group(2)+'\n')
            if detected=='':
              detected=reg.group(2)
          if reg.group(1)=="-1":
            raise ScannerError,reg.group(2)
      return level,detected,virlist

class trophie(sophie):
  '''
  trophie realscanner
  
  It is an alias for sophie. See sophie documentation.
  '''
  name='trophie()'



'''
kav module - Kaspersky support, version 0.2

(c) 2003-2004 Laszlo BALINT (blaci) <balint_laszlo(at)tiszanet.hu>
              Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>
                                                                                
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

'''

from avlib import *
import os,re,string

__all__=['kav','kavclient']

class kav(ascanner):
  '''
  Kaspersky realscanner
  
  Usage: kav(command)

  Where: command is a array which defines command with parameters.
           By default: ['/opt/kav/bin/kavscanner','-j3']

  This scanners is a realscanner, which can be used to scan
  for viruses. It parses Kaspersky antivirus output.
  '''
  name='kav()'
  def __init__(self,command=['/opt/kav/bin/kavscanner','-j3']):
      self.command=command
  def create_args(self,files,dir=''):
      self.args=[]
      for fn in files:
        if dir:
          self.args.append(safe.fn(os.path.join(dir,fn)))
        else:
          self.args.append(safe.fn(fn))
  def scanfile(self,files,dir=''):
      level,detected,ret=0.0,'',[]
      self.create_args(files,dir)
      pf=popen(self.command+self.args)
      pf.tocmd.close()
      rc=pf.wait()
      if rc==1:
        raise ScannerError,'Unable to connect to aveserver.'
      elif rc==9:
        raise ScannerError,'Some of the required parameters are missing from the command line.'
      # read lines from command
      while 1:
        line=pf.readline()
        if line=='': break
        re1=re.search("(INF|SUSP)(ECTED|ICION) (.*)",line)
        if re1:
          debug.echo(3,"INFECTED: kav(): ",line.rstrip())
          if re1.group(1)=="INF": # if it is an indentified virus
            detected=re1.group(3)
            level+=1.0
          else:
            level+=0.75
            if detected=='': # if it is suspected, but no ident. virus
              detected=re1.group(3)
          ret.append(line)
      pf.fromcmd.close()
      debug.echo(4,"KAV: RET: ",[[rc]],ret)
      return level,detected,ret

class kavclient(kav):
  '''
  Kaspersky antivirus client realscanner
  
  Usage: kavclient([socket_path [,command]])

  Where: socket_path is a string, which defines path co aveserver socket
           inside chroot. (default: /var/run/aveserver)
         command is a string, which defines command to run.
           (default: /opt/kav/bin/aveclient)

  This scanners is a realscanner, which can be used to scan
  for viruses. It parses Kaspersky antivirus client output.
  '''
  name='kavclient()'
  def __init__(self,socket_path='/var/run/aveserver',command='/opt/kav/bin/aveclient'):
      self.command=[command,'-s','-p']
      self.command_lastarg=socket_path
  def create_args(self,files,dir=''):
      self.args=[safe.fn(self.command_lastarg)]
      for fn in files:
        if dir:
          self.args.append(safe.fn(os.path.join(dir,fn)))
        else:
          self.args.append(safe.fn(fn))



'''
filter module, version 0.1

(c) 2003 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>
                                                                                
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

'''

from avlib import *
import re

__all__=['filter']

class filter(ascanner):
  '''
  Filter scanner
  
  Usage: filter(['/path/command_name','-options',...])
  
  Where: in [] is a path and argumets of filter binary

  This scanner can be used to filter email through an external program.
  It's return status is always clean.
  '''
  is_spamscan=1
  name='filter()'
  def __init__(self,arg,clean_status=[0]):
      self.arg=arg
      self.clean_status=clean_status
  def scanstream(self,stream):
      pf=popen(self.arg)
      pf.write(stream)
      pf.tocmd.close()
      while True:
        try:
          data=pf.read()
          pf.wait()
          if pf.exitstatus in self.clean_status:
            mail.data=data
            mail.findbody()
          break
        except IOError,(ec,es):
          if ec==4: #Interrupted system call
            continue
          debug.echo(1,"filter(): an IOError occured during data read: ",es)
          pf.wait()
      debug.echo(4,"filter(): exit status: ",pf.exitstatus)
      return 0.0,'',[]



'''
example.py - An example service for sagator.

(c) 2004 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

'''

import re
from aglib import *

__all__=['srv_example']

class srv_example(service):
  '''
  Example service for sagator.
  
  This service can be used as an template to write your own service.
  
  Usage: srv_example(scanners, parameters)
  
  Where: scanners is an array of scanners (see README.scanners for more info)
         parameters is ...

  Example: srv_example(SCANNERS, parameters)
  '''
  
  # Define this service name here. Service name must end with ().
  name='srv_example()'
  
  # This method is called from sagator's config file.
  # You can define some parameters, which are required for this service.
  def __init__(self,scanners,parameters='Default'):
      # You can only store these parameters. This code can't do
      # anything else. For more examples see service class from aglib.py.
      self.SCANNERS=scanners
      self.parameters=parameters
  def main(self):
      # This method is called from sagator's main process.
      # You can fork some processes here and then this method must
      # return.
      # Return value: array of forked pids
      return []
  def accept(self,connects=0):
      # accept method is required only if you are using preforking core
      # from default service (default service.main() function).

      # accept connection
      conn,addr = self.s.accept()
      # you need to reinit scanners before a scan
      for scnr in self.SCANNERS:
        scnr.reinit()
      # also need to reset global class
      globals.reset()
      
      # you need to store data into mail class
      mail.data='........'
      # now you can use checkvir() function to check for viruses
      stat,virname=checkvir(self.SCANNERS)
      # and test it's return values
      if stat==S_OK:
        # It's clean.
        pass
      elif stat==S_FORCE_SEND:
        # It contains a virus, but it's requested to send this email
        # to recipients.
        pass
      elif stat==S_REJECT:
        # You need to reject this email.
        pass
      elif stat==S_DROP:
        # You need to drop this email.
        pass
      elif stat==S_TEMPFAIL:
        # An temporary error occured, send 4XX error code over SMTP.
        pass
      # now you need to safely shutdown this connection
      try:
        conn.shutdown(2)
      except socket.error:
        pass
      # and close this connection
      conn.close()
  # Following methods are not required. They are called from default
  # service main method.
  def fork(self):
      # fork a new process
      pass
  def sighup(self,sn,stack):
      # HUP signal received
      pass
  def sigchld(self,sn,stack):
      # an child exited
      pass
  def sigterm(self,sn,stack):
      # TERM signal received
      pass



'''
Basic interscanners for sagator, version 0.6.1

(c) 2003-2004 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
    
'''

from avlib import *
import string

__all__=['match_all','alternatives','match_any','recover','nothing']

#####################################################################
### match classes

def get_only_scanners(scanners):
    ret=[]
    for scanner in scanners:
      if type(scanner) in [type(()),type([])]:
        ret.extend(get_only_scanners(scanner))
      else:
        try:
          if scanner.is_scanner:
            ret.append(scanner)
        except:
          pass
    return ret

class match_class(interscanner):
  '''parent class for all match clases'''
  def __init__(self,*scanners):
      self.scanners=get_only_scanners(scanners)
      self.rename(self.scanners)
  def destroy(self):
      for scnr in self.scanners:
        scnr.destroy()
  def param(self,p):
      n=0
      for scnr in self.scanners:
        n+=scnr.param(p)
      return n
  def reinit(self):
      for scnr in self.scanners:
        scnr.reinit()
  def get(self,var):
      try:
        r=eval("self."+var)
        if not r:
          raise AttributeError
      except AttributeError:
        for scnr in self.scanners:
          try:
            r+=scnr.get(var)
          except NameError:
            r=scnr.get(var)
          except AttributeError:
            r=''
      return r

class match_all(match_class):
  '''
  match_all() interscanner
  
  Usage: match_all(scanner1(),scanner2(),...)
  
  Where: scanner1(),... are scanners, which must find a virus
  
  Returns a virus only if all scanners returns a virus.
  '''
  name='match_all()'
  def scanstream(self,stream):
      level,vir,ret=0.0,'',[]
      for self.scanner in self.scanners:
        self.scanner.prescan()
        l,v,r=self.scanner.scanstream(stream)
        self.scanner.postscan(l,v,r)
        if v:
          if vir:
            vir+=','+v
          else:
            vir=v
          ret+=r
          level+=l
        else:
          return 0.0,'',[]
      return level,vir,ret

class alternatives(match_class):
  '''
  alternatives() interscanner
  
  Usage: alternatives(scanner1(),scanner2(),...)
  
  Where: scanner1(),... are scanners, which may find a virus
  
  Returns a virus, if one scanner returns a virus.
  Scannig is ended if one scanner returns virus or clean.
  Next scanner from the list is used, if previous scannes fails
  (raises an error). Raises an ScannerError, if all scanners failed.
  '''
  name='alternatives()'
  def scanstream(self,stream):
      level,vir,ret=0.0,'',[]
      for self.scanner in self.scanners:
        try:
          self.scanner.prescan()
          level,vir,ret=self.scanner.scanstream(stream)
          self.scanner.postscan(level,vir,ret)
          return level,vir,ret
        except:
          debug.echo(3,self.name,": scanner ",self.scanner.name," failed: ",
                string.rstrip(debug.traceback_value_str()))
      debug.echo(3,self.name,": All scanners failed")
      raise ScannerError,''

class match_any(match_class):
  '''
  match_any() interscanner
  
  Usage: match_any(scanner1(),scanner2(),...)
  
  Where: scanner1(),... are scanners, which may find a virus
  
  Returns a virus, if one scanner returns a virus.
  Next scanner from list is used, if previous scanner returns
  no virus, or failed (raised an error).
  '''
  name='match_any()'
  def scanstream(self,stream):
      level,vir,ret=0.0,'',[]
      for self.scanner in self.scanners:
        try:
          self.scanner.prescan()
          level,vir,ret=self.scanner.scanstream(stream)
          self.scanner.postscan(level,vir,ret)
          if is_infected(level):
            return level,vir,ret
        except:
          debug.echo(3,self.name,": scanner ",self.scanner.name," failed: ",
                string.rstrip(debug.traceback_value_str()))
          raise
      return 0.0,'',[]

class recover(match_any):
  '''
  recover() interscanner
  
  Usage: recover(scanner1(),scanner2(),...)
  
  Where: scanner1... are scanners
  
  This scanner can be used to recover from scanner fail. If an scanner
  fails (raises an error), this scanner return no virus.
  '''
  name='recover()'
  def scanstream(self,stream):
      level,vir,ret=0.0,'',[]
      for self.scanner in self.scanners:
        try:
          self.scanner.prescan()
          level,vir,ret=self.scanner.scanstream(stream)
          self.scanner.postscan(level,vir,ret)
          if is_infected(level):
            return level,vir,ret
        except:
          debug.echo(3,self.name,": scanner ",self.scanner.name," failed: ",
                string.rstrip(debug.traceback_value_str()))
      return 0.0,'',[]

class nothing(match_any):
  '''
  nothing interscanner
  
  Usage: nothing([some parameters...])
  
  Where: [some parameters...] are ignored :)
  
  This scanner does nothing. :-)
  '''
  name='N()'
  def scanstream(self,stream):
      return match_any.scanstream(self,stream)



'''
Header interscanners for sagator, version 0.6.2

(c) 2003-2004 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
    
'''

from avlib import *
import re
from match import match_any

__all__=['add_header','modify_header','modify_subject','remove_headers']

class add_header(match_any):
  '''
  add_header() - add email header

  usage: add_header('Header_name','header_data',scanners)
     or: add_header(...).onlyif('virname')
  
  where: 'Header_name' is a string, which defines name of the header
         'header_data' is a string, which defines header's content
         'virname' is a string, which defines any virus name.
           By default all virus names at least 1 character long: '.'
         There are some special pattern, which are replaced:
           %V ... virus name
           %S ... scanner name
           %L ... detected level

  This scanner can be used to add an header line to email.
  '''
  name='add_header()'
  def __init__(self,hdr,data,*scanners):
      self.ONLY_IF='.'
      self.HDR=hdr
      self.DATA=data
      match_any.__init__(self,scanners)
  def onlyif(self,pattern='.'):
      self.ONLY_IF=pattern
      return self
  def hdr_replace(self,to,level,detected):
      return re.sub("%V",detected,
               re.sub("%S",globals.found_by.name,
                 re.sub("%L",str(level),
                   to)))
  def scanstream(self,stream):
      level,detected,virlist=match_any.scanstream(self,stream)
      if re.search(self.ONLY_IF,detected,re.IGNORECASE):
        mail.addheader(self.HDR,self.hdr_replace(self.DATA,level,detected))
      return level,detected,virlist

class modify_header(add_header):
  '''
  modify_header() - modifying email headers

  usage: modify_header('^Header_regexp:','Header_regexp: prefix',scanners)
     or: modify_header(...).onlyif('virname')
  
  where: '^Header_regexp:' is a string, which defines regular expression
           to find for this header
         'Header_regexp: prefix' is a string, which defines new header
           data
         'virname' is a string, which defines any virus name.
           By default all virus names at least 1 character long: '.'
         There are some special pattern, which are replaced:
           %V ... virus name
           %S ... scanner name
           %L ... detected level

  This scanner can be used to modify email headers.
  '''
  name='modify_header()'
  def __init__(self,frm,to,*scanners):
      self.ONLY_IF='.'
      self.MOD_FRM=frm
      self.MOD_TO=to
      match_any.__init__(self,scanners)
  def scanstream(self,stream):
      level,detected,virlist=match_any.scanstream(self,stream)
      if re.search(self.ONLY_IF,detected,re.IGNORECASE):
        mail.modheader(self.MOD_FRM,self.hdr_replace(self.MOD_TO,level,detected))
      return level,detected,virlist

class modify_subject(modify_header):
  '''
  modify_subject() - modifying email Subject

  usage: modify_subject('prefix',scanners)
     or: modify_subject(...).onlyif('virname')
  
  where: 'prefix' is an prefix added into Subject.
         'virname' is a string, which defines any virus name.
           By default all virus names at least 1 character long: '.'
         For more information see modify_header() interscanner.

  This scanner can be used to modify email Subject.
  For example original Subject line is:
    Subject: Testing mail
  and i use:
    modify_subject('[SPAM]', spamassassind(['localhost',783]))
  When spam will be found, header will be modifyed to:
    Subject: [SPAM] Testing mail
  '''
  name='modify_subject()'
  def __init__(self,prefix,*scanners):
      modify_header.__init__(self,"^Subject:","Subject: "+prefix,scanners)

class remove_headers(ascanner):
  '''
  remove_headers() - removing email headers

  usage: remove_headers('header name' [,'next header name' [, ...] ])
  
  where: 'header name' is a string, which defines header key to delete

  This scanner can be used to remove email headers.
  '''
  name='remove_headers()'
  def __init__(self,*hdrs):
      self.DEL_HDRS=hdrs
  def scanstream(self,stream):
      for hdr in self.DEL_HDRS:
        mail.delheader(hdr)
      return 0.0,'',[]



'''
config helper for sagator
 
(c) 2003-2004 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>
                                                                                
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
'''

# All scanners included here

from av import *
from as import *
from interscan import *
from stats import status

def SIMPLE_SCANNERS(av,as,
    notify_admin=[],
    notify_sender=False,
    notify_recipients=False,
    deliverto=[],
    quarantine=False,
    quarantine_dir='/tmp/quarantine',
    admin_msg=report.MSG_TMPL,
    recipients_msg=report.MSG_TMPL
  ):
  if notify_sender:
    drop_pattern='(worm|@mm)'
  else:
    drop_pattern='.'
  if notify_recipients:
    my_report_recipients=report_recipients
  else:
    my_report_recipients=nothing
  if quarantine:
    my_quarantine=quarantine
  else:
    my_quarantine=nothing
  return [\
    status("Virus",
      deliver_to(deliverto,
        report(notify_admin,admin_msg,
          my_report_recipients(recipients_msg,
            my_quarantine(quarantine_dir,'',
              drop(drop_pattern,
                av
              )
            )
          )
        )
      )
    ),
    status("Spam",
      deliver_to(deliverto,
        report(notify_admin,admin_msg,
          my_report_recipients(recipients_msg,
            my_quarantine(quarantine_dir,'',
              drop(drop_pattern,
                as
              )
            )
          )
        )
      )
    )
  ]



'''
proxy.py - HTTP proxy daemon service for sagator.

(c) 2004 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

'''

import socket,re,string,os,sys
from cStringIO import StringIO
from aglib import *

__all__=['http_proxy']

BUF_SIZE=16386

HTTP_REPORT='''\
HTTP/1.1 450 Virus Found
Server: SAGATOR
Connection: close
Content-Type: text/html; charset=iso-8859-2

<html><head>
<title>450 Virus Found</title>
</head><body style="background-color: red">
<h1 style="color: yellow">Virus Found !!!</h1>
<p style="color: white">
URL: %s<br/>
Virus name: %s [%f, %s]<br/>
Antivir output: %s<br/>
</p>
</body></html>
'''

HTTP_ERROR='''\
HTTP/1.1 %03d %s
Server: SAGATOR
Connection: close
Content-Type: text/html; charset=iso-8859-2

<html><head>
<title>%03d %s</title>
</head><body>
<h1>%s</h1>
<p>%s</p>
</body></html>
'''

METHODS='(GET|HEAD|POST|PUT|DELETE)'
HOST_PORT='([a-zA-Z0-9.-]+)(:[0-9]+|)'
URI='(http|ftp)://'+HOST_PORT+'/([^ ]*)'
DIR='(/[^ ]*)'
HTTP_VER='HTTP/([0-9]+\.[0-9]+)'

reg_http_whost=re.compile('^'+METHODS+' +'+URI+' +'+HTTP_VER+'\r?$').search
reg_http_nohost=re.compile('^'+METHODS+' +'+DIR+' +'+HTTP_VER+'\r?$').search
reg_hostport=re.compile(HOST_PORT).search
reg_hdr_line=re.compile('^([!-9;-~]+): *(.+?)\r?$',re.M).search
reg_chunk=re.compile('^([0-9A-Fa-f]+)(;.*?|)\r?$').search
reg_response=re.compile('^'+HTTP_VER+' +([0-9]{3}) +(.*)$').search

class http_proxy(service):
  '''
  HTTP proxy service for sagator.
  
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  !!!  WARNING! This service is experimental! Use at your risk!  !!!
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  
  This service can be used as filtering HTTP proxy.
  
  Usage: http_proxy(scanners, host, port, [,prefork])
  
  Where: scanners is an array of scanners (see README.scanners for more info)
         host is a an ip address to bind
         port is a port to bind
         prefork is a number, which defines preforked process count
  
  Example: http_proxy(SCANNERS, '127.0.0.1', 3128)
  
  Warning! Do not forget to block access to this port for non-local users.
  '''
  name='http_proxy()'
  def http_error(self,code=500,s='Internal Server Error',desc=''):
      self.conn.sendall(HTTP_ERROR % (code,s,code,s,s,desc))
      self.conn.shutdown(2)
  def scandata(self):
      globals.reset()
      for scnr in self.SCANNERS:
        level,virname,ret=scnr.scanstream(mail.data)
        scnr.destroy()
        if is_infected(level):
          debug.echo(3,'%s: STATUS: %s [%f,%s]' % \
            (self.name,virname,level,globals.found_by.name))
          break
      if is_infected(level):
        self.conn.sendall(HTTP_REPORT % \
          (mail.sender,virname,level,globals.found_by.name,str(ret)))
        return False
      else:
        return True
  def accept(self,connects=0):
      if connects==0:
        self.conn,self.addr=self.s.accept()
        self.conn.settimeout(300)
        self.connf=self.conn.makefile('rw',0)
      else:
        debug.echo(7,"Persistent connection: ",connects)
        #self.conn.shutdown(2) # PERSISTENT CONNECTIONS ARE NOT WORKING ???
      req=self.connf.readline()
      # Connection closed? Last request?
      if not req:
        return
      debug.echo(3,self.name+": "+req.strip())
      r_http=reg_http_whost(req)
      if r_http:
        method,proto,host,port,path,version_s=r_http.groups()
        if port:
          port=int(port[1:])
        else:
          port=80
      else:
        r_http=reg_http_nohost(req)
        if r_http:
          method,path,version_s=r_http.groups()
          proto='http'
          proto,host,port=('http',None,80)
        else:
          debug.echo(1,self.name+": BAD REQUEST! [%s]" % req.strip())
          self.http_error(desc="Bad Request")
          return
      # method,proto,host,port,path,version defined now
      version=float(version_s)
      # Receive request header
      hdr,phdr='',{'connection':'', 'proxy-connection':''}
      while True:
        l=self.connf.readline()
        if not l:
          debug.echo(1,self.name+": client connection closed by peer")
          self.conn.shutdown(2)
          return
        if (l=='\r') or (l=='\r\n'):
          # end of request header
          break
        hdr+=l
        r=reg_hdr_line(l)
        if r:
          phdr[r.group(1).lower()]=r.group(2)
      # is there an Host: in header
      add_into_header=''
      if phdr.has_key('host'):
        r=reg_hostport(phdr['host'])
        if r:
          host,port=r.groups()
        if port:
          port=int(port)
        else:
          port=80
      else:
        if port==80:
          add_into_header+='Host: %s\r\n' % host
        else:
          add_into_header+='Host: %s:%d\r\n' % (host,port)
      # Is host defined?
      if not host:
        debug.echo(1,self.name+": BAD REQUEST! [%s]" % req.strip())
        self.http_error(desc="Bad Request, host not defined!")
        return
      # check for content length and transfer encoding
      content_length=0
      transfer_encoding=None
      if method.upper()=='POST':
        if phdr.has_key('content-length'):
          content_length=int(phdr['content-length'])
        if phdr.has_key('transfer-encoding'):
          transfer_encoding=phdr['transfer-encoding'].strip()
      # download this page
      newreq='%s /%s HTTP/%s\r\n%s%s\r\n' % \
        (method,path,version_s,hdr,add_into_header)
      debug.echo(6,[req,newreq])
      try:
        cli=socket.socket(socket.AF_INET,socket.SOCK_STREAM)
        cli.settimeout(180)
        cli.connect((host,port))
      except socket.error,(ec,es):
        self.http_error(desc=es)
        return
      if content_length>0:
        post_data=self.conn.recv(content_length)
        debug.echo(4,self.name+": POST DATA: "+str(post_data))
      else:
        post_data=''
      cli.send(newreq+post_data)
      cf=cli.makefile('rw',0)
      # read reply
      while True:
        rhdr=cf.readline()
        r_response=reg_response(rhdr.strip())
        if not r_response:
          debug.echo(1,self.name+": Wrong response: %s " % rhdr.strip())
        resp_code=r_response.group(2)
        debug.echo(4,"Response: ",rhdr.strip())
        # read reply header
        preply_hdr={'connection':'', 'content-length':'0',
                    'transfer-encoding':'' }
        while True:
          l=cf.readline()
          rhdr+=l
          if (l=='\r') or (l=='\r\n'):
            # end of reply header
            break
          r=reg_hdr_line(l)
          if r:
            preply_hdr[r.group(1).lower()]=r.group(2)
        if resp_code=="100": # 100 Continue
          self.conn.sendall(rhdr)
        else:
          break
      content_length=int(preply_hdr['content-length'])
      debug.echo(6,"reply_header: ",rhdr)
      # it is encoded as chunked?
      if preply_hdr['transfer-encoding'][:7]=='chunked':
        data=StringIO()
        while True:
          l=cf.readline()
          # connection closed?
          if not l: break
          # next chunk?
          r_chunk=reg_chunk(l.strip())
          if r_chunk:
            chunk_size=int(r_chunk.group(1),16)
            debug.echo(7,'chunk: ',chunk_size)
            if chunk_size==0:
              cli.shutdown(2)
              break
            data.write(cf.read(chunk_size))
            # read a plain CRLF
            cf.readline()
          else:
            debug.echo(1,self.name+": ERROR! Chunk expected! [%s]" % l.strip())
            break
        mail.data=data.getvalue()
        mail.sender="%s://%s:%d/%s" % (proto,host,port,path)
        mail.recip=[self.addr[0],str(self.addr[1])]
        del data
        if self.scandata(): # clean?
          self.conn.sendall("%s%X\r\n" % (rhdr,len(mail.data)))
          self.conn.sendall(mail.data)
        # accept next REQUEST if connection was not closed
        if l:
          if preply_hdr['connection']!="close":
            if phdr['proxy-connection']!="close":
              self.accept(connects+1)
      else:
        data=StringIO()
        if preply_hdr['connection']=="close":
          while True:
            d=cli.recv(BUF_SIZE)
            if not d: # connection closed
              break
            data.write(d)
        else:
          while data.tell()<content_length:
            data.write(cli.recv(content_length-data.tell()))
        mail.data=data.getvalue()
        del data
        mail.sender="%s://%s:%d/%s" % (proto,host,port,path)
        mail.recip=[self.addr[0],str(self.addr[1])]
        if self.scandata(): # clean?
          self.conn.sendall(rhdr+mail.data)
        if preply_hdr['connection']!="close":
          if phdr['proxy-connection']!="close":
            # persistent connection
            self.accept(connects+1)
      debug.echo(6,"closing connection")
      self.conn.shutdown(2)



'''
dspam module, version 0.1

(c) 2003 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>
                                                                                
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

'''

from avlib import *
import socket,re,string
import pwd,grp,os

try:
  import pydspam

  DSR_ISSPAM = pydspam.DSR_ISSPAM
  DSR_ISINNOCENT = pydspam.DSR_ISINNOCENT

  DSS_ERROR = pydspam.DSS_ERROR
  DSS_CORPUS = pydspam.DSS_CORPUS
  DSS_INOCULATION = pydspam.DSS_INOCULATION

  DST_TEFT = pydspam.DST_TEFT
  DST_TOE = pydspam.DST_TOE
  DST_TUM = pydspam.DST_TUM
  DST_NOTRAIN = pydspam.DST_NOTRAIN
  
  DSF_CHAINED = pydspam.DSF_CHAINED
  DSF_SIGNATURE = pydspam.DSF_SIGNATURE
  DSF_COPYBACK = pydspam.DSF_COPYBACK
  DSF_NOISE = pydspam.DSF_NOISE
  DSF_WHITELIST = pydspam.DSF_WHITELIST
  DSF_MERGED = pydspam.DSF_MERGED

  class dspam(ascanner):
    '''
    dspam realscanner - uses dspam library
    
    Usage: dspam(home,user,group,classify,source,train,flags)
    
    Where: home is dspam home dir ('/var/dspam')
           user is dspam user (user running sagator's process)
           group is dspam group (group running sagator's process)
           train is training mode (TEFT)
           flagh are dspam flags (DSF_SIGNATURE)

    For detailed description see DSPAM documentation.

    Examples:
      basic usage:
        dspam()
      training emails badly assigned as spams:
        drop('',dspam(classify=DSR_ISSPAM,source=DSS_ERROR,train=DST_TOE))
      training emails badly assinged as inoocent:
        drop('',dspam(classify=DSR_ISINNOCENT,source=DSS_ERROR,train=DST_TOE))
    '''
    name='dspam()'
    OP=pydspam.DSM_PROCESS
    FLAGS=pydspam.DSF_CHAINED|pydspam.DSF_NOISE
    CLASSIFICATION=pydspam.DSR_NONE
    SOURCE=pydspam.DSS_NONE
    TRAINING_MODE=DST_TEFT
    def __init__(self,home='/var/dspam',user=None,group=None,classify=None,source=None,train=None,flags=pydspam.DSF_SIGNATURE):
        self.FLAGS |= flags
        if not user:
          user=pwd.getpwuid(os.getuid())[0]
        if not group:
          group=grp.getgrgid(os.getgid())[0]
        self.DSPAM=pydspam.dspam(home,op=self.OP,flags=self.FLAGS,user=user,group=group)
        if classify:
          self.CLASSIFICATION=classify
        if self.CLASSIFICATION!=pydspam.DSR_NONE:
          self.DSPAM.classification(self.CLASSIFICATION)
        if source:
          self.SOURCE=source
        if self.SOURCE!=pydspam.DSS_NONE:
          self.DSPAM.source(self.SOURCE)
        if train:
          self.TRAINING_MODE=train
          self.DSPAM.training_mode(train)
    def __del__(self):
        self.DSPAM.__del__()
    def param(self,p):
        if p=='--dspam-source-corpus':
          self.SOURCE=pydspam.DSS_CORPUS
          self.DSPAM.source(self.SOURCE)
        elif p=='--dspam-source-error':
          self.SOURCE=pydspam.DSS_ERROR
          self.DSPAM.source(self.SOURCE)
        elif p=='--dspam-source-innoculation':
          self.SOURCE=pydspam.DSS_INOCULATION
          self.DSPAM.source(self.SOURCE)
        elif p=='--dspam-classify-spam':
          self.CLASSIFICATION=pydspam.DSR_ISSPAM
          self.DSPAM.classification(self.CLASSIFICATION)
        elif p=='--dspam-classify-innocent':
          self.CLASSIFICATION=pydspam.DSR_ISINNOCENT
          self.DSPAM.classification(self.CLASSIFICATION)
        elif p=='--dspam-training-teft':
          self.TRAINING_MODE=pydspam.DST_TEFT
          self.DSPAM.training_mode(pydspam.DST_TEFT)
        elif p=='--dspam-training-toe':
          self.TRAINING_MODE=pydspam.DST_TOE
          self.DSPAM.training_mode(pydspam.DST_TOE)
        elif p=='--dspam-training-tum':
          self.TRAINING_MODE=pydspam.DST_TUM
          self.DSPAM.training_mode(pydspam.DST_TUM)
        elif p=='--dspam-training-notrain':
          self.TRAINING_MODE=pydspam.DST_NOTRAIN
          self.DSPAM.training_mode(pydspam.DST_NOTRAIN)
        else:
          return 0
        return 1
    def scanstream(self,stream):
        debug.echo(4,self.name+": settings: DST=%s, DSS=%s, DSR=%s" % \
          (pydspam.DST[self.TRAINING_MODE],
           pydspam.DSS[self.SOURCE],
           pydspam.DSR[self.CLASSIFICATION]))
        sig=''
        if self.SOURCE in [pydspam.DSS_ERROR,pydspam.DSS_INOCULATION]:
          # try to get signature from meesage header
          reg_sig=re.compile('^X-DSPAM-Signature: *([a-z0-9]+)\r?$',re.I|re.M).search(stream)
          if reg_sig:
            sig=reg_sig.group(1)
            self.DSPAM.get_signature(sig)
          else:
            debug.echo(1,self.name+": ERROR: Can't get signature ID!")
        # process through dspam
        r,p,c=self.DSPAM.process(stream)
        # save signature, if it is made
        if self.CLASSIFICATION==pydspam.DSR_NONE:
          sig=self.DSPAM.save_signature()
        # simulate dspam output
        ret=['result='+str(r),'probability='+str(p),'confidence='+str(c),
             "sig='%s'" % str(sig)]
        debug.echo(4,self.name+": "+', '.join(ret))
        if self.CLASSIFICATION==pydspam.DSR_ISINNOCENT:
          return 0.0,'',['innocence classified']
        if self.CLASSIFICATION==pydspam.DSR_ISSPAM:
          return 1.0,'SPAM',['spam classified']
        mail.addheader('X-DSPAM-Confidence',str(c))
        mail.addheader('X-DSPAM-Probability',str(p))
        if r==pydspam.DSR_ISSPAM:
          mail.addheader('X-DSPAM-Result','Spam')
          mail.addheader('X-DSPAM-Signature',sig)
          return 1.0+p,'SPAM',ret
        elif r==pydspam.DSR_ISINNOCENT:
          mail.addheader('X-DSPAM-Result','Innocent')
          return p,'',ret
        debug.echo(3,self.name+": WARNING: Unknown return from dspam!")
        return 0.0,'',[]
  class dspam_classify(dspam):
    '''
    dspam_classify() realscanner
    
    This scanner can be used to classify messages as spams/hams.
    For parameters see dspam() scanner documentation.
    
    Example:
      configure SCANNERS=[ dspam_classify(source=DSS_CORPUS,flags=0) ]
      and you can run sgscan to learn messages from a box:
      Training ham (clean messages):
        sgscan --dspam-training-teft --dspam-source-corpus \
               --dspam-classify-innocent MAILBOX
      Training spam:
        sgscan --dspam-training-teft --dspam-source-corpus \
               --dspam-classify-spam MAILBOX
    '''
    name='dspam_classify()'
    #OP=pydspam.DSM_CLASSIFY
    #FLAGS=pydspam.DSF_CHAINED|pydspam.DSF_NOISE
    CLASSIFICATION=pydspam.DSR_ISSPAM
    SOURCE=pydspam.DSS_CORPUS
except ImportError,es:
  class dspam(ascanner):
    '''
    This method is undefined, because pydspam module is not found.
    Please recompile sagator, if you really need it.
    '''
    name='dspam()'
    errorstring="Importing of pydspammodule failed!"
    def __init__(self,*arg1,**arg2):
        pass
    def scanstream(self,stream):
        debug.echo(1,self.name+": scanstream(): "+self.errorstring," [",es,"]")
        raise ScannerError,self.errorstring
    def scanfile(self,files,dirname=''):
        debug.echo(1,self.name+": scanstream(): "+self.errorstring," [",es,"]")
        raise ScannerError,self.errorstring
  class dspam_classify(dspam):
    name='dspam_classify()'



'''
decompress module, version 0.3.2

(c) 2003-2004 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>
                                                                                
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

'''

from avlib import *
import re,string,os
import filetype,resource

__all__=['decompress','decompressors','internal_decompressors']

decompressors={
  'zip':	[[ 'unzip','-P','s','-o' ]],
  'rar':	[[ 'unrar','x','-p-','-y'], ['rar','x','-p-','-y' ]],
  'ace':	[[ 'unace','x','-y' ]],
  'arj':	[[ 'unarj','x','-y'], [ 'arj','x','-y' ]],
  'zoo':	[[ 'unzoo','-x','-j'], [ 'zoo','-x','-j' ]],
  'lha':	[[ 'lha','xf' ]],
  'tar':	[[ 'tar','xvf' ]],
  #'tar.gz':	[[ 'tar','xvzf' ], [ 'gzip','-d' ]],
  #'tgz':	[[ 'tar','xvzf' ]],
  #'tar.bz2':	[[ 'tar','xvjf' ], [ 'bzip2','-d' ]],
  'bz2':	[[ 'bzip','-d','-k' ]],
  'gz':		[[ 'gzip','-d' ]],
}

internal_decompressors={}

# ZIP support
try:
  import zipfile
  class i_unzip:
    '''UnZip file'''
    def __init__(self,maxfilesize=100*1024*1024):
        '''
        if decompressed file is larger than maxfilesize, it is skipped
        '''
        self.maxfilesize=maxfilesize
    def unpack(self,zipfilename,dir):
        debug.echo(4,"Internal unzip: ",zipfilename)
        filenames,fsize=[],0L
        zip=zipfile.ZipFile(safe.fn(zipfilename),'r')
        for i in zip.namelist():
          info=zip.getinfo(i)
          if (info.file_size<self.maxfilesize) & (info.file_size>=0):
            debug.echo(5,"Decompressing ... ",i)
            tf=mktemp('','-'+normalize_filename(i),'w',0600)
            filenames.append(os.path.join(dir,tf.name))
            tf.f.write(zip.read(i))
            fsize=fsize+tf.f.tell()
            tf.f.close()
          else:
            debug.echo(0,"Internal UnZip: ERROR: FileSize too large: ",info.file_size)
        zip.close()
        return filenames,fsize
  internal_decompressors['zip']=i_unzip()
except ImportError:
  class zipfile:
    BadZipfile='BadZipFile'

# GZIP support
try:
  import gzip
  class i_ungzip:
    '''unpack gzip file'''
    def unpack(self,filename,dir):
        debug.echo(4,"Internal ungzip: ",filename)
        gzf=gzip.GzipFile(safe.fn(filename),'r')
        fn_noext=re.sub('\.gz$','',re.sub('\.tgz$','.tar',filename))
        tf=mktemp('',normalize_filename(fn_noext),'w',0600).autorm()
        fsize=0L
        while 1:
          buf=gzf.read(BUFSIZE)
          if buf:
            tf.f.write(buf)
            fsize=fsize+len(buf)
          else:
            break
        tf.f.close()
        gzf.close()
        return [os.path.join(dir,fn_noext)],fsize
  internal_decompressors['gz']=i_ungzip()
except ImportError:
  pass

class decompress(interscanner):
  '''
  decompress scanner - used to decompress zip,rar,arj,zoo,tar,...
  
  Usage: decompress(filescanner1(),[a,b,c,d])
  
  Where: filescanner1() is an filescanner, which can scan all files
         [a,b,c,d] is an array of 4 numbers:
           a = max number of recursions
           b = max number of decompressed files
           c = max disk usage with all of decompressed files
           d = max address space usage
  
  It is a scanner, which can decompress some of files.
  You can define your own decompressors and assing it to any extensions.
  This scanner is a filescanner only, because external decompressors
  can unpack only files. A used scanner must be also a filescanner.
  If not, use file2stream() interscanner.
  
  Filetype is determined on file content (not on file extension).
  It can detect some exe SFX archives too.
  '''
  name='decompress()'
  def __init__(self,scanner,max_limit=[3,500,50l*1024*1024,20l*1024*1024]):
      self.name=self.name[:-1]+scanner.name+')'
      self.scanner=scanner
      self.max_recursion=max_limit[0]
      self.max_file_count=max_limit[1]
      self.max_file_size=max_limit[2]
      self.max_as_limit=max_limit[3]
      self.initvars()
  def initvars(self):
      self.vir,self.ret='',[]
      self.defiles=[]
      self.rmfiles=[]
      self.fsize=0
      self.fcount=0
      self.showed_error=0
  def destroy(self):
      self.scanner.destroy()
      if debug.debug_level<10:
        rdel(self.rmfiles)
      else:
        debug.echo(4,"decompress(): unfinished destroy:",self.rmfiles)
      self.initvars()
  def scanfile(self,files,dir='',level=0):
      # check for max_recursion
      if level>=self.max_recursion:
        if self.showed_error<1:
          debug.echo(0,"Decompress: ERROR: Max recursion reached, ", level)
          self.showed_error=1
        return 0.0,'',[]
      # check for max_filesize
      if self.fsize>=self.max_file_size:
        if self.showed_error<2:
          debug.echo(0,"Decompress: ERROR: Max file size reached, ", self.fsize)
          self.showed_error=2
        return 0.0,'',[]
      # check for max_file_count
      if self.fcount>=self.max_file_count:
        if self.showed_error<2:
          debug.echo(0,"Decompress: ERROR: Max file count reached, ", self.fcount)
          self.showed_error=2
        return 0.0,'',[]
      # decompress files ...
      for fname in files:
        try:
          nfname=normalize_filename(fname)
          ft=filetype.file(fname)[0]
        except IndexError:
          ft=""
        debug.echo(6,"decompress(): file name/type: ",nfname," / ",ft)
        already_scanned=0
        if ft in internal_decompressors:
          try:
            zipdir=mktemp(fname,'.dir','d',0700)
            self.rmfiles.append(zipdir.name)
            os.chdir(safe.fn(zipdir.name))
            newfiles,size=internal_decompressors[ft].unpack(fname,zipdir.name)
            self.fsize=self.fsize+size
            self.fcount=self.fcount+len(newfiles)
            self.scanfile(newfiles,dir,level+1)
            already_scanned=1
          except zipfile.BadZipfile:
            debug.echo(2,"Internal decompressor: ERROR: file: ",nfname,
                  " -> ",debug.traceback_value_str())
          except:
            debug.echo(0,"Internal decompressor: ERROR: file: ",nfname,
                  " -> ",debug.traceback_value_str())
            debug.traceback(4,"internal_decompressor: ")
        if (already_scanned<1) and (ft in decompressors.keys()):
          progs=decompressors[ft]
          zipdir=mktemp(fname,'.dir','d',0700)
          self.rmfiles.append(zipdir.name)
          for arg in progs:
            try:
              pf=popen(arg+[safe.fn(fname)],safe.fn(zipdir.name),
                {resource.RLIMIT_FSIZE:self.max_file_size,
                 resource.RLIMIT_AS:self.max_as_limit})
              decerr=pf.readlines()
              if pf.wait()>0:
                # add fname to decompressed files if unpack returned error
                self.defiles.append(fname)
              if pf.execerror=='':
                break
              else:
                debug.echo(0,"Decompressor: ERROR: code: ",pf.exitstatus)
            except:
              debug.echo(3,'Decompressor failed: ',str(arg+[fname]))
              debug.traceback(3,"Decompressor: ")
          newfiles,size=rlistdir(zipdir.name,zipdir.name)
          self.fsize=self.fsize+size
          self.fcount=self.fcount+len(newfiles)
          self.scanfile(newfiles,dir,level+1)
        else: # no decompressor found for this filetype, add it as is
          self.defiles.append(fname)
      if level==0:
        debug.echo(5,"decompressed files: ",[self.defiles])
        self.scanner.prescan()
        level,vir,ret=self.scanner.scanfile(self.defiles,dir)
        self.scanner.postscan(level,vir,ret)
      else:
        level,vir,ret=0.0,'',[]
      return level,vir,ret



#!/usr/bin/python2

'''
milter.py - milter service for sagator

(c) 2003-2004 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

'''

from aglib import *
from StringIO import StringIO
import signal
import sys,time

__all__=['milter']

try:
  import Milter

  class sgMilter(Milter.Milter):
    "Sagator's milter class."

    def log(self,msg,level=2):
        debug.echo(level,"milter(%s):" % self.CID,msg)

    def connect(self,hostname,unused,hostaddr):
        self.CID=randomchars(6,'0123456789abcdef')
        self.log("Connection from: "+hostname+" at "+\
                 time.strftime("%c",time.localtime()))
        self.comm1="XFORWARD ADDR=%s NAME=%s HELO= PROTO=SMTP\r\n" % (hostaddr,hostname)
        self.mail=mail_class()
        return Milter.CONTINUE

    def hello(self,hostname):
        self.comm1+="HELO %s\r\n" % hostname
        return Milter.CONTINUE

    def envfrom(self,f,*s):
        self.log('MAIL FROM: '+f,2)
        self.t0=time.time()
        self.mail.__init__() # reinit, may be a new email
        self.mail.comm=self.comm1+"MAIL FROM: %s\r\n" % f
        self.mail.sender=f
        return Milter.CONTINUE

    def envrcpt(self,to,*s):
        self.mail.comm+="RCPT TO: "+to+"\r\n"
        self.log('RCPT TO: '+to,2)
        try:
          self.mail.recip.append(rfc822.parseaddr(to)[1])
        except:
          pass
        return Milter.CONTINUE

    def header(self,name,val):
        msg=name+": "+val+"\r\n"
        self.mail.df.write(msg)
        return Milter.CONTINUE

    def eoh(self):
        self.mail.df.write("\r\n")
        return Milter.CONTINUE

    def body(self,chunk):
        self.mail.df.write(chunk)
        return Milter.CONTINUE

    def eom(self):
        self.mail.comm+="DATA\r\n"
        # reinit mail class and store mail data
        mail.__init__()
        mail.df=self.mail.df
        mail.comm=self.mail.comm
        mail.sender=self.mail.sender
        mail.recip=self.mail.recip
        mail.close()
        v,virname=checkvir(self.SCANNERS)
        self.tw=time.time()-self.t0
        debug.echo(2,"STATS: "+str(self.tw)+" seconds, ",
                     len(mail.data)," bytes, status: ",virname)
        if (v==S_OK) | (v==S_FORCE_SEND):
          self.log("ACCEPT: 250 Ok",1)
          stats.update(len(mail.data),self.tw)
          return Milter.ACCEPT
        elif v==S_REJECT:
          msg="550 Content rejected - VIRUS "+virname;
          self.log("REJECT: "+msg,1)
          stats.update(len(mail.data),self.tw,1)
          self.setreply(msg[0:3],None,msg[4:])
          return Milter.REJECT
        elif v==S_DROP:
          self.log("DROP: 250 mail dropped, VIRUS "+virname,1)
          stats.update(len(mail.data),self.tw,1)
          return Milter.DISCARD
        self.log("TEMPFAIL: 451 "+str(virname),1)
        stats.update(0,0,0,1) # update fail statistics
        return Milter.TEMPFAIL

    def close(self):
        self.log("Connection closed.")
        return Milter.CONTINUE

    def abort(self):
        self.log("Connection aborted!")
        return Milter.CONTINUE

  def milterInit(scanners,name,conn):
      sgMilter.SCANNERS=scanners
      Milter.factory = sgMilter
      Milter.set_flags(Milter.CHGBODY + Milter.CHGHDRS + Milter.ADDHDRS)
      Milter.runmilter(name,conn,256)

except ImportError,es:
  def milterInit(scanners,name,conn):
      debug.echo(0,"ImportError: Can't import Milter module!")
      debug.echo(0,"Install sendmail milters in python from:")
      debug.echo(0,"  http://www.bmsi.com/python/milter.html")
      sys.exit(1)

class milter(service):
  '''
  Milter service for sagator.
  
  This service can be used to start sagator as milter filter.
  
  Usage: milter(scanners,name,connection)
  
  Example: milter(SCANNERS,"sagator","inet:3333@127.0.0.1")
  
  For more information about milter's parameters see milter documentation.
  
  You need python's milter module to run this service:
    http://www.bmsi.com/python/milter.html
  '''
  name='milter()'
  def __init__(self,scanners,name,conn):
      self.SCANNERS=scanners
      self.m_name=name
      self.m_conn=conn
      self.EXITING=False
      self.childs=[]
  def main(self):
      test_scanners(self.SCANNERS)
      # chroot
      dochroot()
      # and now start the daemon
      signal.signal(signal.SIGCHLD,self.sigchld)
      signal.signal(signal.SIGHUP,sighup)
      signal.signal(signal.SIGTERM,self.sigterm)
      pid=self.daemonize()
      if pid>0:
        return [pid]
      self.PID=os.getpid()
      self.fork()
      # wait for child exit
      while True:
        time.sleep(60)
      sigterm('Terminating') # kill self
      sys.exit(0)
      return [] # required by pychecker
  def fork(self):
      if self.EXITING:
        return -1
      if self.childs!=[]:
        return -1
      p=os.fork()
      if p==0:
        debug.echo(1,"milter(): service started, waiting for connections ...")
        milterInit(self.SCANNERS,self.m_name,self.m_conn)
        debug.echo(1,"milter(): Exiting ...")
        sys.exit(0)
      else:
        self.childs.append(p)
      return p



'''
rlimit.py - An "service" to set resource limits in sagator.

(c) 2004 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

'''

from aglib import *

__all__=['rlimit']

class rlimit(service):
  '''
  rlimit service for sagator.
  
  This service can be used to set resource limits for sagator.
  
  Usage: rlimit(PARAM1=value1, PARAM2=value1, ...)
  
  Where: PARAM1,... are resource parameter names
         value1,... are resource values

  Example: rlimit(AS=30000000)
  '''
  name='rlimit()'
  def __init__(self,**args):
      self.ARGS=args
  def main(self):
      import resource
      debug.echo(1,"rlimit(): LIMITS:",str(self.ARGS))
      for res,limit in self.ARGS.items():
        try:
          eres=eval('resource.RLIMIT_'+res)
          resource.setrlimit(eres,(limit,limit))
          debug.echo(7,"rlimit(): Limit %s set to %s"
            % (res,str(resource.getrlimit(eres))))
        except:
          debug.echo(1,"rlimit(): ERROR: Can't set limit for resource: "+res)
          raise
      return []



'''
Antivir Gateway Library for Sagator

(c) 2003-2004 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>

This program is a email antivir gate for an SMTP server.

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

'''

import os,sys,string,re,time
import pwd,grp,socket,rfc822

from avlib import *
import stats

TEST_MSG='''\
From: Test User <test@tester.salstar.sk>\r
To: Test User2 <test2@tester2.salstar.sk>\r
Date: Fri, 16 Apr 2004 15:38:51 +0200 (CEST)\r
Subject: testing mail\r
\r
simple mail\r
'''

#########################################################################
### service class template

class service:
  '''Sagator service template'''
  name='ASrv()'
  def __init__(self,scanners,host,port,prefork=2,listen=5):
      self.SCANNERS=scanners
      self.BINDTO=(host,port)
      self.LISTEN=listen
      self.PREFORK=prefork
      self.EXITING=False
  def daemonize(self):
      # redirect stout and strerr
      debug.dup()
      # fork
      pid=os.fork()
      if (pid<=0) and (globals.daemon):
        globals.pidf.close()
      return pid
  def main(self):
      test_scanners(self.SCANNERS)
      # bind to socket
      self.s=socket.socket(socket.AF_INET,socket.SOCK_STREAM)
      self.s.setsockopt(socket.SOL_SOCKET,socket.SO_REUSEADDR,1)
      try:
        self.s.bind(self.BINDTO)
        self.s.listen(self.LISTEN)
      except socket.gaierror,(ec,es):
        debug.echo(0,self.name+": ERROR: BIND ",es)
        sys.exit(1)
      # and now start the daemon
      signal.signal(signal.SIGCHLD,self.sigchld)
      signal.signal(signal.SIGHUP,sighup)
      signal.signal(signal.SIGTERM,signal.SIG_DFL)
      pid=self.daemonize()
      if pid>0:
        self.s.close()
        signal.signal(signal.SIGCHLD,signal.SIG_DFL)
        return [pid]
      # chroot
      dochroot()
      debug.echo(1,self.name+": service started, waiting for connections ...")
      # prefork
      signal.signal(signal.SIGTERM,self.sigterm)
      signal.signal(signal.SIGHUP,self.sighup)
      self.PID=os.getpid()
      self.childs=[]
      for i in range(self.PREFORK):
        self.fork()
      debug.echo(3,self.name+": pids: ",str(self.childs))
      try:
        while True:
          time.sleep(60)
      except KeyboardInterrupt:
        self.sigterm(None,None) # kill childs
        raise
  def accept(self,connects=0):
      debug.echo(0,self.name+": This method must be overwrtten in subclass!")
      sys.exit(1)
  def fork(self):
      try:
        pid=os.fork()
        if pid>0:
          self.childs.append(pid)
          return pid
      except OSError,(ec,es):
        debug.echo(1,"ERROR: Can't fork:",es)
        return -1
      signal.signal(signal.SIGCHLD,signal.SIG_DFL)
      signal.signal(signal.SIGHUP,sighup) # reopen log
      signal.signal(signal.SIGTERM,signal.SIG_DFL)
      self.time1,self.timec='',1
      while True:
        try:
          self.accept()
        except SmtpcError,(ses,es):
          # error is written, accept new connection
          pass
        except socket.error,err:
          try:
            ec=err[0]
            es=err[1]
          except IndexError:
            ec=0
            es=str(err)
          if ec==107: # Transport endpoint is not connected
            debug.echo(1,self.name+": Connection closed by peer.")
          elif ec==104: # Connection reset by peer
            debug.echo(1,self.name+": Connection reset by peer.")
          elif ec==32: # 'Broken pipe'
            debug.echo(0,self.name+": ERROR: Broken pipe.")
          elif ec!=4:
            debug.echo(0,self.name+': ERROR: SocketError: ',es)
            debug.traceback(4,self.name+": ")
        except KeyboardInterrupt:
          self.sigterm(None,None)
        except MemoryError:
          debug.echo(1,self.name+": ERROR: OUT OF MEMORY! EXITING PROCESS!")
          # it's safe to exit process now
          sys.exit(1)
        except:
          debug.echo(1,self.name+": ERROR: Unknown error!")
          debug.traceback(1,self.name+": ")
      return 0 # child process
  def sighup(self,sn,stack):
      # reopen log for master process
      sighup(sn,stack)
      # reopen logs for all childs
      for pid in self.childs:
        try:
          os.kill(pid,signal.SIGHUP)
        except:
          pass
  def sigchld(self,sn,stack):
      # try to remove older zombies (sigchld lost?)
      try:
        while 1:
          pid,s=os.waitpid(0,os.WNOHANG)
          if pid==0:
            break
          if pid in self.childs:
            self.childs.remove(pid)
          debug.echo(1,self.name+": Child exited ... pid,status: ",(pid,s))
      except OSError: # no child processes?
        pass
      if not self.EXITING:
        # start another child
        debug.echo(1,self.name+": WARNING: An child exited? Starting new one ...")
        self.fork()
  def sigterm(self,sn,stack):
      self.EXITING=True
      if os.getpid()!=self.PID:
        debug.echo(1,self.name[:-2]+"(child): Exiting - SIGTERM ...")
        os._exit(0)
      # kill all childs
      for pid in self.childs:
        debug.echo(5,self.name[:-2]+'(master): killing child %d (TERM signal)' % pid)
        try:
          os.kill(pid,signal.SIGTERM)
        except OSError:
          pass
      time.sleep(0.1)
      if self.childs:
        time.sleep(1)
        for pid in self.childs:
          debug.echo(3,self.name[:-2]+'(master): killing child %d (KILL signal)' % pid)
          try:
            os.kill(pid,signal.SIGKILL)
          except OSError:
            pass
      os._exit(0)

#########################################################################
### Some helpful functions

# chroot, setgid and setuid
def dochroot():
    '''
    Do chroot() and setuid(),setgid()
    '''
    def func_err(fn,f):
        try:
          eval(f)
          return 0
        except OSError,(ec,es):
          debug.echo(0,"WARNING! Can't do %10s: %s" % (fn,es))
        except AttributeError:
          debug.echo(0,"WARNING! Can't do "+f+": Your python don't have this function!")
        return 1
    uid,gid=0,0
    if globals.USER:
      try:
        uid=pwd.getpwnam(globals.USER)[2]
      except KeyError,es:
        debug.echo(0,"ERROR, getpwnam: ",[[globals.USER]],es)
        sys.exit(100)
    if globals.GROUP:
      try:
        gid=grp.getgrnam(globals.GROUP)[2]
      except KeyError,es:
        debug.echo(0,"ERROR, getgrnam: ",[[globals.GROUP]],es)
        sys.exit(100)
    if globals.CHROOT:
      if func_err('chroot',"os.chroot('"+globals.CHROOT+"')"):
        safe.ROOT_PATH=globals.CHROOT
    if gid:
      func_err('setgroups',"os.setgroups(["+str(gid)+"])")
      func_err('setgid',"os.setgid("+str(gid)+")")
    if uid:
      func_err('setuid',"os.setuid("+str(uid)+")")
    if (os.getuid()==0) or (os.getgid()==0
        or (os.geteuid()==0) or (os.getegid()==0)):
      debug.echo(0,"ERROR! I can't setuid and i am running as root!")
      sys.exit(10)

def test_scanners(scanners):
    for scnr in scanners:
      scnr.reinit()
    if os.fork()==0:
      dochroot()
      globals.scan_only=True
      scanarr,scannames=[],'None'
      mail.data=TEST_MSG
      for scnr in scanners:
        if scnr.name:
          debug.echo(1, "Testing "+scnr.name+"...")
        level,detected,virlist,scan_report,err=do_scan(scnr)
        if scan_report:
          debug.echo(0, "Scanner ",scnr.name," test failed! Disable it manually!")
          debug.echo(0, "  ",scan_report)
        scanarr=scanarr+[scnr.name]
      mail.data=''
      sys.exit(0)
    os.wait()

def do_scan(scnr,fn=''):
    '''
    Scan email (stored in mail class) for viruses.
    Use scnr scanner.
    '''
    level,detected,virlist=0.0,'',[]
    try:
      if fn:
        scnr.filename=fn
      scnr.prescan()
      level,detected,virlist=scnr.scanstream(mail.data)
      scnr.postscan(level,detected,virlist)
      scnr.destroy()
      return level,detected,virlist,'',None
    except socket.error,(ec,es):
      debug.traceback(4,"do_scan: ")
      return level,detected,virlist,\
        scnr.name+": SocketError: "+es,'Socket error: '+es
    except (IOError,OSError):
      debug.traceback(4,"do_scan: ")
      return level,detected,virlist,\
        scnr.name+": Error: "+debug.traceback_value_str(),'Error: '+debug.traceback_value_str()
    except (ScannerError,ValueError),es:
      debug.traceback(4,"do_scan: ")
      return level,detected,virlist,\
        scnr.name+": "+str(es),'Error: '+str(es)
    except KeyboardInterrupt:
      raise
    except:
      debug.traceback(4,"do_scan: ")
      return level,detected,virlist,scnr.name+": unknown error",'Unknown error'

def do_param(p,scanners):
    n=0
    for scnr in scanners:
      n+=scnr.param(p)
    return n

def checkvir(scanners):
    '''
    Check email (stored in mail class) for viruses.
    '''
    try:
      level,detected,virlist=0.0,'',[]
      scanarr,scannames=[],'None'
      for scnr in scanners:
        l,detected,virlist,scan_reply,err=do_scan(scnr)
        level+=l
        if scan_reply:
          debug.echo(0, "CheckVir: ERROR: ",scan_reply)
          return S_TEMPFAIL,scan_reply
        else:
          scanarr=scanarr+[scnr.name]
          scannames=string.join(scanarr,' ')
        if is_infected(level,detected):
          debug.echo(3,"Status: ",[detected,virlist,level])
          break
      mail.addvirusinfo(detected,scannames)
      if not is_infected(level,detected):
        return S_OK,'CLEAN' # mail delivered OK
      else: # VIRUS FOUND
        rstat=globals.ACTION
        debug.echo(5,"action: ",rstat)
        return rstat,detected
    except socket.error,(ec,es):
      debug.echo(0,"checkvir: ERROR: socket.error: ",[[ec]],es)
      debug.traceback(4,"checkvir: ")
      return S_TEMPFAIL,es
    except IOError,(ec,es):
      debug.echo(0,"checkvir: ERROR: IO: ",es)
      debug.traceback(4,"checkvir: ")
      return S_TEMPFAIL,'IOError'
    except:
      debug.traceback(3,"checkvir: ")
      return S_TEMPFAIL,'AnyError, see sagator logs'

#########################################################################
### Signal handlers

def sigchld(sn,stack):
    ''' signal: Child process terminated '''
    # try to remove older zombies (sigchld lost?)
    try:
      while 1:
        pid,s=os.waitpid(0,os.WNOHANG)
        if pid==0:
          break
        debug.echo(1,"Child exited ... pid,status: ",(pid,s))
    except OSError: # no child processes?
      pass

def sighup(sn,stack):
    ''' signal: HUP signal '''
    debug.echo(3,"SIGHUP: Reopening log ...")
    debug.reopen()

def sigusr2(sn,stack):
    ''' signal: USR2 signal '''
    debug.echo(1,"SIGUSR2: Tracelog request received ...")
    debug.debug_level=debug.trace_level
    debug.stack(1,"SIGUSR2: ")

def sigterm(sn,stack=None):
    ''' signal: Terminate process '''
    if type(sn)==type(''):
      debug.echo(1,"SERVER: Exiting - %s ..." % sn)
    else:
      debug.echo(1,"SERVER: Exiting - SIGTERM ...")
    try:
      os.unlink(globals.pid_file)
    except OSError:
      pass
    try:
      if globals.stat_coll.pid>0:
        os.kill(globals.stat_coll.pid,signal.SIGTERM)
    except:
      pass
    sigchld(None,None) # unzombie all childs
    sys.exit(0)



'''
Command realscanners for sagator, version 0.5

(c) 2003-2004 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
    
'''

from avlib import *
import re,string

__all__=['cmd','cmd_clamav','cmd_kavscanner','cmd_drweb','cmd_fprot',
         'cmd_trendmicro','cmd_bdc','cmd_uvscan','cmd_vbuster',
         'cmd_mks']

#####################################################################
### exec_any class

class cmd(ascanner):
  '''
  cmd() realscanner, used to scan over command line programs
  
  Usage: cmd(['command','arg1',...],StatusCLEAN,StatusINFECTED,RE)
  
  Where: 'command' is a string, which command to run
         'argX' are command line arguments, as last argument the
           filename is added
         StatusCLEAN may be:
           an array of integers, which defines "clean" exit codes
           an regexp, indication no virus found
         StatusINFECTED may be:
           an array of integers, which defines "virus found" exit codes
         RE is an regular expression string, which can extract virus
           names from scanner output.
  '''
  name='cmd()'
  DIR=('DIR')
  FILES=('FILES')
  def __init__(self,args,clean_status=[0],infected_status=[1],vreg=None):
      self.args=args
      if type(clean_status)==type(''):
        self.clean_status=re.compile(clean_status,re.M)
      else:
        self.clean_status=clean_status
      if type(infected_status)==type(''):
        self.infected_status=re.compile(infected_status,re.M)
      else:
        self.infected_status=infected_status
      if type(vreg)==type(''):
        self.vreg=re.compile(vreg,re.M)
      else:
        self.vreg=vreg
  def check_vname(self,ret):
      reg1=self.vreg.search(string.join(ret,''))
      if reg1:
        return reg1.group(1)
      else:
        return "UnknownVirus"
  def scanfile(self,files,dir=''):
      args=[]
      for arg in self.args:
        if arg==self.DIR:
          args.append(safe.fn(dir))
        elif arg==self.FILES:
          for f in files:
            args.append(safe.fn(f))
        else:
          args.append(arg)
      pf=popen(args)
      pf.tocmd.close()
      ret=pf.readlines()
      debug.echo(6,ret)
      pf.wait()
      pf.fromcmd.close()
      if type(self.infected_status)==type([]):
        if pf.exitstatus in self.infected_status:
          return 1.0,self.check_vname(ret),ret
      else:
        regi=self.infected_status.search(string.join(ret,''))
        if regi:
          return 1.0,self.check_vname(ret),ret
      if type(self.clean_status)==type([]):
        if pf.exitstatus in self.clean_status:
          return 0.0,'',[]
      else:
        regc=self.clean_status.search(string.join(ret,''))
        if regc:
          return 0.0,'',[]
      debug.echo(1,self.name+': command: ',string.join(args),
                 ' ExitStatus: ',str(pf.exitstatus),ret)
      raise ScannerError,self.name+' error'

class cmd_clamav(cmd):
  '''
  cmd_clamav() scanner uses clamscan command from clamav project.
  
  See http://www.clamav.net/ for more info.
  
  Usage: cmd_clamav(['command_name_with_full_path'])
  '''
  name='cmd_clamav()'
  def __init__(self,command='clamscan'):
      cmd.__init__(self,
        [command,'--stdout','--disable-summary','-r',cmd.DIR],
        [0],[1],
        '^.*?: (?!Infected Archive)(.*) FOUND$'
      )

class cmd_kavscanner(cmd):
  '''
  cmd_kavscanner() scanner uses kavscanner command from KasperskyLab
    kavscanner.
  
  See http://www.kaspersky.com/ for more info.
  
  Usage: cmd_kavscanner(['command_name_with_full_path'])
  '''
  name='cmd_kavscanner()'
  def __init__(self,command='/opt/kav/bin/kavscanner'):
      cmd.__init__(self,
        [command,'-i1','-xp',cmd.DIR],
        [0,10,15], [5,20,21,25],
        '(?:CURED|INFECTED|CUREFAILED|WARNING|SUSPICION) (.*)'
      )

class cmd_drweb(cmd):
  '''
  cmd_drweb() scanner uses drweb command from DRWEB antivirus.
  
  See http://www.sald.com/, http://drweb.imshop.de/ for more info.
  
  Usage: cmd_drweb(['command_name_with_full_path'])
  '''
  name='cmd_drweb()'
  def __init__(self,command='/usr/local/drweb/drweb'):
      cmd.__init__(self,
        [command,'-path',cmd.DIR,'-al','-go','-ot','-cn','-upn','-ok-'],
        [0,32], [1,9,33], 
        ' infected (?:with|by)(?: virus)? (.*)$'
      )

class cmd_fprot(cmd):
  '''
  cmd_fprot() scanner uses f-prot command from FRISK F-Prot antivirus.
  
  See http://www.f-prot.com/ for more info.
  
  Usage: cmd_fprot(['command_name_with_full_path'])
  '''
  name='cmd_fprot()'
  def __init__(self,command='f-prot'):
      cmd.__init__(self,
        [command,'-dumb','-archive','-packed',cmd.DIR],
        [0,8], [3,6],
        'Infection: (.+)'
      )

class cmd_trendmicro(cmd):
  '''
  cmd_trendmicro() scanner uses trendmicro command from
    Trend Micro FileScanner.
  
  See http://www.trendmicro.com/ for more info.
  
  Usage: cmd_trendmicro(['command_name_with_full_path'])
  '''
  name='cmd_trendmicro()'
  def __init__(self,command='/etc/iscan/vscan'):
      cmd.__init__(self,
        [command,'-za','-a',cmd.DIR],
        [0], 'Found virus',
        'Found virus (.+) in'
      )

class cmd_bdc(cmd):
  '''
  cmd_bdc() scanner uses bdc command from BitDefender antivirus.
  
  See http://www.bitdefender.com/ for more info.
  
  Usage: cmd_bdc(['command_name_with_full_path'])
  '''
  name='cmd_bdc()'
  def __init__(self,command='bdc'):
      cmd.__init__(self,
        [command,'--all','--arc','--mail',cmd.DIR],
        '^Infected files *:0+(?!\d)',
          '^(?:Infected files|Identified viruses|Suspect files) *:0*[1-9]',
        '(?:suspected|infected): (.*)(?:\033|$)'
      )

class cmd_uvscan(cmd):
  '''
  cmd_uvscan() scanner uses uvscan command from McAfee AntiVirus.
  
  See http://www.mcafee.com/ for more info.
  
  Usage: cmd_uvscan(['command_name_with_full_path'])
  '''
  name='cmd_uvscan()'
  def __init__(self,command='uvscan'):
      cmd.__init__(self,
        [command,'--secure','-rv','--mime','--summary','--noboot','-',cmd.DIR],
        [0], [13],
        ' the (.+) (?:virus|trojan)'
      )

class cmd_vbuster(cmd):
  '''
  cmd_vbuster() scanner uses vbuster command from VirusBuster.
  
  See http://www.virusbuster.hu/en/ for more info.
  
  Usage: cmd_vbuster(['command_name_with_full_path'])
  '''
  name='cmd_vbuster()'
  def __init__(self,command='vbscan'):
      cmd.__init__(self,
        [command,'--action=skip',cmd.DIR],
        [0,1,2,3,4,5], '(?:virus|iworm) found: ',
        "(?:virus|iworm) found: (.+) \.\.\. "
      )

class cmd_mks(cmd):
  '''
  cmd_mks() scanner uses mks antivirus
  
  See http://linux.mks.com.pl/ for more info.
  
  Usage: cmd_mks(['command_name_with_full_path' [,extra_params_array]])
  '''
  name='cmd_mks()'
  def __init__(self,command='mks32',params=[]):
      cmd.__init__(self,
        [command]+params+['--',cmd.FILES],
        [0], [1,2,4],
        "^mks_vir: \t--(?:|destroyed by |scanstring of )virus (.+)"
      )



#!/usr/bin/python

import sys,os,re
LISTEN=10
MAX_PROCS=10
from default import *
from config import *

fn=os.path.join(os.path.dirname(sys.argv[0]),'config.py')
cfg=open(fn,'r').read()
cfg_orig=cfg

# add SRV[] array
if SRV==[]:
  out=""
  out+="from srv import *\n";
  out+="SRV=[\n"
  out+="  # statistics collector\n"
  try:
    out+="  collector('"+COLLECTOR_SERVER[0]+\
         "', "+str(COLLECTOR_SERVER[1])+"),\n"
  except NameError:
    out+="  #collector('0.0.0.0',28),\n"
  out+="  # SMTP daemon (for posfix etc.)\n"
  try:
    out+="  smtpd(SCANNERS,'"+BIND_ADDR[0]+"', "+str(BIND_ADDR[1])+")\n"
  except (NameError,TypeError):
    out+="  #smtpd(SCANNERS,'127.0.0.1',27)\n"
  out+="  # Milter daemon\n"
  try:
    out+="  milter(SCANNERS,'"+MILTER[0]+"', "+MILTER[1]+")\n"
  except (NameError,TypeError):
    out+='  #milter(SCANNERS,"sagator","inet:3333@127.0.0.1")\n'
  out+="  # Filte\n  #avfilter(SCANNERS)\n"
  out+="]\n"
  print "Updating %s by:" % fn
  print out
  cfg+="\n"+out

# remove LISTEN, MAX_PROCS, BIND_ADDR, ...
cfg0=cfg
cfg=re.compile('^ *(LISTEN|MAX_PROCS|BIND_ADDR|COLLECTOR_SERVER) *=.*?$\n',re.MULTILINE
      ).sub('',cfg)
if cfg0!=cfg:
  print "Removing obsolete parameters."

# update from 0.5.3-rc7 to 0.5.3-rc8
cfg0=cfg
cfg=re.compile('(smtpd|milter)\(([\'"])',
      re.MULTILINE).sub('\\1(SCANNERS,\\2',cfg)
cfg=re.compile('avfilter\(\)',
      re.MULTILINE).sub('avfilter(SCANNERS)',cfg)
if cfg0!=cfg:
  print "Updating SERVICES for new syntax."
reg_srv=re.search('^ *SRV[ \n]*=[ \n]*\[.*?^ *\]',cfg,re.M|re.DOTALL)
reg_sca=re.search('^ *SCANNERS[ \n]*=[ \n]*\[.*?^ *\]',cfg,re.M|re.DOTALL)
try:
  if reg_srv.start()<reg_sca.start():
    cfg=cfg[:reg_srv.start()] + '\n' \
       +cfg[reg_sca.start():] + '\n' \
       +cfg[reg_srv.start():reg_srv.end()]
    print "Moving SRV array after SCANNERS (SRV uses SCANNERS)."
except:
  pass # update not required, at least one array not defined

if len(sys.argv)>1:
  print cfg
  sys.exit(1)

if cfg==cfg_orig:
  #print "No update needed."
  sys.exit(0)
else:
  # store old config
  open(fn+'.sagatorsave','w').write(cfg_orig)
  print 'Your old configuration is saved into "%s".' % (fn+'.sagatorsave')
  # save, close and rename file
  open(fn,'w').write(cfg)
